/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file features.cpp
@ingroup ExamplePrograms
@brief Obtains several image features from input video frames

This application different workers which obtain four different image features, depicting on four respective image canvas the resulting features. These obtained image features are:
 - Canny edges.
 - Harris corners.
 - Maximal stable extremal regions (MSER).
 - Thresholded contours.

@section Usage Usage of the program.
Once installed the QVision, compile using the <i>qmake</i> and <i>make</i> tools:
@code # qmake; make @endcode
Execute the application with the following line:
@code ./features --URL=<source video identifier string> @endcode
For example:
@code ./features --URL=http://perception.inf.um.es/public_data/videos/misc/minuto.avi @endcode
See section @ref QVMPlayerURLFormat for further info about the format and types of source video identifier strings.

@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <qvdta.h>
#include <qvip.h>

#include <QVApplication>
#include <QVMPlayerCamera>
#include <QVDefaultGUI>
#include <QVImageCanvas>
#include <QVPolyline>
#include <QVHarrisPointDetector>
#include <QVCannyEdgeDetector>
#include <QVMSERDetector>
#include <QVNumericPlot>
#include <QVHistogramPlot>

#ifndef DOXYGEN_IGNORE_THIS

class MyWorker: public QVWorker
	{
	public:
		MyWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,1> >("Output image", outputFlag);
			addProperty<int>("Max pixel", outputFlag);
			addProperty<int>("Min pixel", outputFlag);
			addProperty<QList<double> >("MinMaxList", outputFlag);
			addProperty<QList<double> >("FirstRow", outputFlag);
			}

		void iterate()
			{
			QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");

			uchar min, max;
			Max(image, max);
			Min(image, min);
			setPropertyValue<int>("Max pixel", max);
			setPropertyValue<int>("Min pixel", min);
			QList<double> minmaxlist;
			minmaxlist << min << max;
			setPropertyValue<QList<double> >("MinMaxList", minmaxlist);

			QList<double> firstrow;
			for (uint i = 0; i < 256 ; i++) firstrow << image(i, 0);
			setPropertyValue<QList<double> >("FirstRow", firstrow);

			QVImage<uChar> dest(image);

			AddC(image, 10, dest);

			setPropertyValue< QVImage<uChar,1> >("Output image", dest);
			}
	};

class ContourExtractorWorker: public QVWorker
	{
	public:
		ContourExtractorWorker(QString name): QVWorker(name)
			{
			addProperty<int>("Threshold", inputFlag,	128, "Threshold for a point to count as pertaining to a region", 0, 255);
			addProperty<int>("MinAreaIPE", inputFlag,	0, "Minimal area to keep points in the IPE algorithm", 0, 50);
			addProperty< QVImage<uChar,3> >("Input image", inputFlag|outputFlag);
			addProperty< QList<QVPolyline> >("Internal contours", outputFlag);
			addProperty< QList<QVPolyline> >("External contours", outputFlag);
			}

		void iterate()
			{
			// 0. Read input parameters
			const QVImage<uChar,1> image = getPropertyValue< QVImage<uChar,3> >("Input image");
			const uInt	rows = image.getRows(), cols = image.getCols(),
					threshold = getPropertyValue< int >("Threshold"),
					minAreaIPE = getPropertyValue< int >("MinAreaIPE");

			timeFlag("Read input parameters");
		
			// 1. Get contours from image
			const QList<QVPolyline> contours = getConnectedSetBorderContoursThreshold(image, threshold);
			timeFlag("Get contours from image");

			// 2. Apply IPE
			QList<QVPolyline> ipeContours;

			foreach(QVPolyline polyline, contours)
				{
				QVPolyline ipePolyline;
				IterativePointElimination(polyline, ipePolyline, minAreaIPE);
				if (ipePolyline.size() > 0)
					ipeContours.append(ipePolyline);
				}

			timeFlag("IPE filtering");

			// 3. Export contours to output property
			QList<QVPolyline> internalContours, externalContours;

			foreach(QVPolyline polyline, ipeContours)
				if (polyline.direction)
					internalContours.append(polyline);
				else
					externalContours.append(polyline);

			setPropertyValue< QList< QVPolyline> >("Internal contours",internalContours);
			setPropertyValue< QList< QVPolyline> >("External contours",externalContours);
			timeFlag("Computed output contours");
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Obtains several features from input video frames."
		);

	ContourExtractorWorker contoursWorker("Contours Extractor Worker");
	QVCannyEdgeDetector cannyWorker("Canny Operator Worker");
	QVHarrisPointDetector cornersWorker("Harris Worker");
	//QVMSERDetector mserWorker("MSER Worker");

	QVMPlayerCamera camera("Video");

	camera.linkProperty(&contoursWorker,"Input image");
	camera.linkProperty(&cannyWorker,"Input image");
	camera.linkProperty(&cornersWorker,"Input image");
	//camera.linkProperty(&mserWorker,"Input image");

	QVDefaultGUI interface;

	QVImageCanvas contourCanvas("Contours");
	contoursWorker.linkProperty("Input image", contourCanvas);
	contoursWorker.linkProperty("Internal contours", contourCanvas);
	contoursWorker.linkProperty("External contours", contourCanvas);
	//contourCanvas.setPropertyValue<QColor>("Color for External contours", Qt::blue);

	QVImageCanvas cannyCanvas("Canny");
	cannyWorker.linkProperty("Output image", cannyCanvas);
	cannyWorker.linkProperty("Output contours", cannyCanvas);

	QVImageCanvas cornersCanvas("Harris");
	cornersWorker.linkProperty("Input image", cornersCanvas);
	cornersWorker.linkProperty("Feature locations", cornersCanvas);

	/*QVImageCanvas mserCanvas("MSER Regions");
	mserWorker.linkProperty("Input image", mserCanvas);
	mserWorker.linkProperty("MSER contours", mserCanvas);*/
	return app.exec();
	}

#endif
