/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file rectify.cpp
@ingroup ExamplePrograms
@brief Rectifies a 2D plane from the input image.

@todo Detailed description should follow.

@section Usage
Once compiled, the program should be used with this command line:
\code
	./rectify --URL=<path or url to video>
\endcode

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#include <qvmath.h>
#include <qvdta.h>
#include <qvip.h>
#include <qvprojective.h>

#include <QVApplication>
#include <QVMPlayerCamera>
#include <QVDefaultGUI>
#include <QVImageCanvas>
#include <QVMatrix>

#include <QVImageCanvas>
#include <QVMatrix>
#include <QVWorker>

#include <hessiancornersextractor.h>
#include <culebrillasdisplayer.h>
#include <featuretracker.h>
#include <planardetector.h>

//#define GRAB_MOUSE

#ifndef DOXYGEN_IGNORE_THIS

/////////////////////////////////

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h> 
#include <X11/cursorfont.h>
#include <X11/extensions/XTest.h>

////////////////////////////////////////////////////////////
// Worker classes

class MouseTracker: public QVWorker
	{
	private:
		QPointF center;
		QVMatrix accum;

		void mouseMove( int x, int y )
			{
			Display *display = XOpenDisplay(NULL);
			int screen = DefaultScreen( display );
			XTestFakeMotionEvent ( display, screen, x, y, 0 );
			XCloseDisplay( display );
			}
		
		void mouseButtonPress( int button, int isPress, unsigned long delay )
			{
			Display *display = XOpenDisplay(NULL);
			int screen = DefaultScreen( display );
			XTestFakeButtonEvent ( display, button , isPress, delay );
			XCloseDisplay( display );
			}

	public:
		MouseTracker(QString name): QVWorker(name), center(0,0), accum(QVMatrix::identity(3))
			{
			// Input properties
			addProperty< double >("Sensitivity", inputFlag, 1, "--", 0.1, 10);
			addProperty< QVImage<uChar> >("Output image", outputFlag);
			addProperty< QList<QPointF> >("Cross", outputFlag);
			addProperty< QVMatrix >("Euclidean transformation", inputFlag);

			addTrigger("Reset pointer");

			QVImage<uChar> img(500,500);
			Set(0, img);
			setPropertyValue< QVImage<uChar> >("Output image", img);
			}

		void processTrigger(QString triggerName)
			{
			center = QPointF(0,0);
			accum = QVMatrix::identity(3);
			}

		void iterate()
			{
			// 0. Read input property values.
			const QVMatrix Hfound = getPropertyValue< QVMatrix >("Euclidean transformation");
			const double sensitivity = getPropertyValue< double >("Sensitivity");

			accum = Hfound * accum;

			QVVector v1(3), v2(3), v3(3);
			v1[0] = 0.1; v1[1] = 0.0; v1[2] = 1.0;
			v2[0] = 0.0; v2[1] = 0.0; v2[2] = 1.0;
			v3[0] = 0.0; v3[1] = 0.1; v3[2] = 1.0;

			QPointF	center1 = sensitivity * 250 * (accum * v1),
				center2 = sensitivity * 250 * (accum * v2),
				center3 = sensitivity * 250 * (accum * v3),
				center1_bis(250 - center1.x(), 250 + center1.y()),
				center2_bis(250 - center2.x(), 250 + center2.y()),
				center3_bis(250 - center3.x(), 250 + center3.y());

			QList<QPointF> cross;
			cross.append(center1_bis);
			cross.append(center2_bis);
			cross.append(center3_bis);

			#ifdef GRAB_MOUSE
				mouseMove(center3.x(), center3.y());
			#else
				setPropertyValue< QList<QPointF> >("Cross", cross);
			#endif
			timeFlag("Export culebrillas as polylines");
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Finds a template pattern on the image."
		);

	QVMPlayerCamera camera("Video");

	// Corner extraction
	HessianCornersExtractor cornersExtractor("Hessian corners extractor");
	camera.linkProperty(&cornersExtractor,"Input image");

	// Feature tracker
	FeatureTracker featureTracker("Feature tracker worker");
	cornersExtractor.linkProperty("Corner responses", &featureTracker, "Feature responses", QVWorker::SynchronousLink);
	cornersExtractor.linkProperty("Corner locations", &featureTracker, "Feature locations", QVWorker::SynchronousLink);

	// Culebrillas processor
	MouseTracker mouseTracker("Mouse tracker");
	featureTracker.linkProperty(&mouseTracker, QVWorker::SynchronousLink);

	#ifndef GRAB_MOUSE

	// Planar detector
	PlanarDetector planarDetector("Planar finder", false);
	featureTracker.linkProperty("Culebrillas container", &planarDetector, "Culebrillas container", QVWorker::AsynchronousLink);

	// Data output
	QVDefaultGUI interface;

	QVImageCanvas imageCanvasPointer("Culebrillas");
	mouseTracker.linkProperty("Output image", imageCanvasPointer);
	mouseTracker.linkProperty("Cross", imageCanvasPointer);
	imageCanvasPointer.setPropertyValue<QColor>("Cross $$ color", Qt::white);

	// Culebrillas
	CulebrillasDisplayer culebrillasDisplayer("Culebrillas processor");
	cornersExtractor.linkProperty("Input image", &culebrillasDisplayer, "Input image", QVWorker::SynchronousLink);
	featureTracker.linkProperty("Culebrillas container", &culebrillasDisplayer, "Culebrillas container", QVWorker::SynchronousLink);
	planarDetector.linkProperty("Planar culebrillas", &culebrillasDisplayer, "Planar culebrillas", QVWorker::AsynchronousLink);

	QVImageCanvas imageCanvas1("Culebrillas");
	culebrillasDisplayer.linkProperty("Input image", imageCanvas1);
	//imageCanvas1.linkProperty(culebrillasDisplayer, "Culebrillas head", Qt::red);
	//imageCanvas1.linkProperty(culebrillasDisplayer, "Planar culebrillas polylines head", Qt::red);

	//imageCanvas1.linkProperty(culebrillasDisplayer, "Planar culebrillas polylines head", Qt::blue);
	culebrillasDisplayer.linkProperty("Culebrillas", imageCanvas1);
	culebrillasDisplayer.linkProperty("Planar culebrillas polylines", imageCanvas1);

	#endif

	return app.exec();
	}

#endif
