/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvworker.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVWORKER_H
#define QVWORKER_H

#include <QStringList>
#include <QThread>
#include <QTime>

#include <qvutils/qvcpustatcontroler.h>
#include <QVPropertyContainer>

/*!
@class QVWorker qvworker.h qvcore/qvworker.h
@brief Base class for Worker threads.

This class can be used to construct worker threads for the Worker Design Pattern. Worker threads classes should inherit from QVWorker, and implement abstract function @ref iterate, with the body of the function code 

For a detailed usage of this class in a QVision application, see section @ref ProgrammingModel.

@ingroup qvcore
*/
class QVWorker: public QThread, public QVPropertyContainer
	{
	Q_OBJECT

	public:
		/// @brief Different states in which a worker object can be.
		typedef enum
			{
			/// Worker is running. @ref iterate() function will be called
			/// continuously, until slots @ref pause() or @ref finish() are 
			/// called.
			Running,
			/// Worker will execute @ref iterate() function once, and then will
			/// change to @ref Paused status.
			RunningOneStep,
			/// Worker is paused. It won't call function @ref iterate(), or
			/// finish execution thread, until slots @ref unPause() or
			/// @ref finish() are called respectively.
			Paused,
			/// Worker is stoped. It won't call function @ref iterate() (but don't
			/// block its conected sychronous workers), or finish execution thread,
			/// until slots @ref finish() are called respectively.
			Stoped,
			/// Worker is finished. Execution thread ended.
			Finished
			} TWorkerStatus;

		/// @brief Constructor for QVWorker class.
		///
		/// Inits worker, and set it's status to @ref Running.
		/// @param name the (optional) name given to the worker.
		QVWorker(const QString name = QString());

		/// @brief Copy Constructor for QVWorker class.
		///
		/// Creates a new worker form a given worker.
		/// @param other the given worker.
		QVWorker(const QVWorker &other);

		/// @brief Destructor for QVWorker class.
		~QVWorker();

		/// @brief Function to be defined in subclasses with processing code.
		///
		/// This function should contain the code to process data, that will be
		/// executed continuously until worker is finished.
		virtual void iterate() { std::cerr << "WARNING: redefinition of worker() is deprecated. Redefine iterate() instead." << std::endl; worker(); };

		/// \deprecated Code to process inputs should be defined in function @ref iterate() instead.
		///
		/// @see iterate()
		virtual void worker() {};

		/// @brief Sets a minimum ms delay for each iteration (default is 0)
		///
		/// This function should be called to stablish a minimum delay for
		/// each iteration (that is, it should be used to limit the speed at
		/// which the worker iterates). Of course, if the real processing
		/// iteration time is longer than the delay set with this function, it
		/// will not have any visible effect.
		///
		/// Warning: due to underlying operating system time management
		/// imprecissions, low values (typically under 40 ms) of this parameter
		/// can result in longer delays than expected.
		void setMinimumDelay(int ms) { minms = ms; };

		/// @brief Function that can be used to check finish status of worker.
		/// @returns true if worker is finished.
		bool isFinished() const { return status == Finished; }

		/// @brief Function that can be used to check paused status of worker.
		/// @returns true if worker is paused.
		bool isPaused() const { return status == Paused; }

		/// @brief Function that can be used to check stopped status of worker.
		/// @returns true if worker is stoped.
		bool isStoped() const { return status == Stoped; }

		/// @brief Function that can be used to check running status of worker.
		/// @returns true if worker is running.
		bool isRunning() const { return status == Running; }

		/// @brief Function to obtain the status of the worker.
		/// @returns status of the worker.
		TWorkerStatus getStatus() const { return status; }
/*
		/// @brief Completely unlinks a property holder.
		///
		/// A worker only can be unlinked if the worker's status is Finished.
		void unlink();
*/
		/// @brief Function to obtain the number of iteration the worker has performed.
		/// @returns number of times that the iterate() function has been called.
		int getIteration() const { return numIterations; }

		/// @brief Function to obtain if the stats are enabled.
		/// @returns if the stats are enabled.
		bool isStatsEnabled() const { return statsEnabled; }

		/// @brief Gets CpuStat object for the worker.
		///
		/// Every worker has a @ref CpuStat type object that mantains cpu time
		/// performance for different parts (code fragments) of function
		/// @ref iterate(). This object is updated in method @ref timeFlag(), when
		/// stablishing a performance breakpoint in the execution of the worker.
		///
		/// If th stats are disabled, return an empty QVStat.
		///
		/// @returns constant reference to worker's CpuStat object.
		QVStat getCpuStat() const {
			if (statsEnabled) return cpuStatControler->value();
			else return QVStat();
		}

		/// @brief Prints CpuStat object for the worker.
		///
		/// Periodically prints performance statistics for the worker.
		/// If the stats are disabled, do nothing.
		///
		/// @param freq The frequency (iterations) whitch prints the stats.
		///
		void setPrintStatsFrequency(int freq) {
			if (statsEnabled) {
				cpuStatControler->setFreq(freq);
				cpuStatControler->setWorkerName(getName());
			}
		}

		/// @brief Prints CpuStat object for the worker.
		///
		/// Prints performance statistics for the worker.
		/// If the stats are disabled, do nothing.
		///
		///
		void printStats()
			{
			if (statsEnabled)
				cpuStatControler->printStats();
			}

		/// @brief Adds a trigger to the worker.
		///
		/// Triggers are just named actions that can be called upon a QVWorker
		/// to change it's state, initialize itself, or whatever action the
		/// programmer considers convenient, in a completely asynchronous
		/// manner.
		/// When added to a worker, they will depict a button in the
		/// corresponding property widget in the interface, that will do a call
		/// to @ref processTrigger() function when pressed.
		///
		/// This method is provided just for convenience. The same effect could
		/// be achieved just by adding a new slot when subclassing QVWorker,
		/// and connecting it to a button clicked signal. But using this
		/// generic function avoids the programmer to have to declare the
		/// new worker class as a Q_OBJECT, if he just needs the trigger
		/// functionality, and does not want to declare any other signals or
		/// slots in his new worker class.
		/// @see processTrigger()
		void addTrigger(QString name)			{ triggerList.push_back(name); }

		/// @brief Returns the list of triggers (as a QStringList) for the worker.
		const QStringList getTriggerList() const	{ return triggerList; }

	public slots:
		/// @brief Set worker status to @ref Paused.
		///
		void pause() { qDebug() << "QVWorker::pause()"; status = Paused; emit statusUpdate(status); }

		/// @brief Set worker status to @ref Running.
		///
		void unPause() { qDebug() << "QVWorker::unPause()"; status = Running; emit statusUpdate(status); }

		/// @brief Set worker status to @ref RunningOneStep.
		///
		void step() { qDebug() << "QVWorker::step()"; status = RunningOneStep; emit statusUpdate(status); }

		/// @brief Finish worker execution.
		///
		void stop() { qDebug() << "QVWorker::stop()"; status = Stoped; emit statusUpdate(status); }

		/// @brief Finish worker execution.
		///
		void finish() { qDebug() << "QVWorker::finish()"; status = Finished; emit statusUpdate(status); }

		/// @brief Process trigger activations in the worker.
		///
		/// Triggers are just named actions that can be called upon a QVWorker
		/// to change it's state, initialize itself, or whatever action the
		/// programmer considers convenient, in a completely asynchronous
		/// manner.
		///
		/// When added to a worker, they will depict a button in the
		/// corresponding property widget in the interface, that will do a call
		/// to @ref processTrigger() function when pressed.
		/// Use @ref addTrigger to add triggers to a worker, and redefine this
		/// function to capture their activations.
		/// @see addTrigger()
		/// @param name the name of the trigger that was activated.
		virtual void processTrigger(QString name) { Q_UNUSED(name); }

	signals:
		/// @brief Signal emited before each call to @ref iterate() function.
		///
		void startIteration();

		/// @brief Signal emited after each call to @ref iterate() function.
		///
		void endIteration(uint id, int iteration);

		/// @brief Signal emited when worker changes its status.
		///
		void statusUpdate(QVWorker::TWorkerStatus);

	private:
		bool statsEnabled;
		QVStatControler *cpuStatControler;
		int numIterations, maxIterations;
		TWorkerStatus status;
		QStringList triggerList;
		QTime iterationTime;
		int curms,minms;

	protected:
		void run();

		/// @brief Stablish a performance breakpoint in the execution of the worker.
		///
		/// This function should be used in the body of @ref iterate() function,
		/// to stablish a performance breakpoint, at which the execution time will be
		/// statistically measured.
		///
		/// If the stats are disabled, do nothing.
		///
		/// These statistics can be plotted through the graphical user interface (see
		/// @ref WorkerWidget section at the @ref QVGUI documentation), creating a
		/// @ref QVCPUPlot object, or to the user console if the parameter
		/// <i>--"print stats"=true</i> was used in the application command line.
		void timeFlag(const QString flag)
			{
			qDebug() << "QVWorker::timeFlag("<< flag <<")";
			if (statsEnabled) cpuStatControler->setFlag(flag);
			}

		void workerIterate();
	};

Q_DECLARE_METATYPE(QVStat);
#endif
