/*
 *	Copyright (C) 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */


#include <QtGui>

#include "itemnode.h"
#include "slatewindow.h"

ItemNode::ItemNode(ItemProperties item, QString _name, SlateWindow *wind, QGraphicsItem * parent, QGraphicsScene * scene): Node(parent, scene),
	itemProp(item), numProp(itemProp.getProperties().size()), window(wind), clickedPoint(-1), externalMarkedPoint(-1), markedPoint(-1)
{
    myTextColor = Qt::darkGreen;
    myOutlineColor = Qt::darkBlue;
    myBackgroundColor = Qt::white;

    setFlags(ItemIsMovable | ItemIsSelectable);


	// obtiene el rectángulo exterior
    QFontMetricsF metrics = qApp->font();
	lineSpacing = metrics.lineSpacing();
    outlinerect = metrics.boundingRect(itemProp.getType());

	const QList<QString> props = itemProp.getProperties();
	for (int i = 0; i < props.size(); i++)
		outlinerect |= metrics.boundingRect(props[i]);

    outlinerect.adjust(0.0, 0.0, lineSpacing*4, props.size()*lineSpacing);


	// obtiene el texto a escribir
	QString nameAndProp(itemProp.getType());
	for (int i = 0; i < props.size(); i++)
		nameAndProp += "\n" + props[i];

	setText(nameAndProp);

	type = itemProp.getType();
	name = _name;
}

ItemNode::~ItemNode()
{
     foreach (Link *link, getLinks())
        delete link;
}

void ItemNode::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget * /* widget */)
{
    QPen pen(myOutlineColor);
    if (option->state & QStyle::State_Selected) {
        pen.setStyle(Qt::DotLine);
        pen.setWidth(2);
    }
    painter->setPen(pen);
    painter->setBrush(myBackgroundColor);

    QRectF rect = outlineRect();
	painter->drawRect(rect);
	painter->drawLine((int)rect.left(), (int)(rect.top() + lineSpacing), (int)rect.right(), (int)(rect.top() + lineSpacing));

    painter->setPen(myTextColor);
    painter->drawText(rect, Qt::AlignTop | Qt::AlignHCenter, myText);

	// y pinta los puntos de unión a ambos lados del texto
    painter->setPen(Qt::white);
	painter->setBrush(Qt::black);
	for (int i = 0; i < numProp; i++)
	{
		if (itemProp.isInput(i)) // si tiene enlace de entrada lo pinto
	    	painter->drawEllipse((int)(rect.left() + lineSpacing*0.25), (int)(rect.top() + lineSpacing*(i + 1.25)), (int)(lineSpacing*0.5), (int)(lineSpacing*0.5));
		if (itemProp.isOutput(i)) // // si tiene enlace de salida lo pinto
			painter->drawEllipse((int)(rect.right() - lineSpacing*0.75), (int)(rect.top() + lineSpacing*(i + 1.25)), (int)(lineSpacing*0.5), (int)(lineSpacing*0.5));
	}

	// si uno de los puntos está marcado, lo pinta dependiendo de su validez
	if (markedPoint >= 0)
	{
		if (markedValidity) painter->setBrush(Qt::green);
		else				painter->setBrush(Qt::red);

		if (markedPoint < numProp)
			painter->drawEllipse((int)(rect.left() + lineSpacing*0.25), (int)(rect.top() + lineSpacing*(markedPoint + 1.25)), (int)(lineSpacing*0.5), (int)(lineSpacing*0.5));
		else
			painter->drawEllipse((int)(rect.right() - lineSpacing*0.75), (int)(rect.top() + lineSpacing*(markedPoint - numProp + 1.25)), (int)(lineSpacing*0.5), (int)(lineSpacing*0.5));
	}
}

// devuelve la posición (en cordenadas de la scena) del centro del punto indicado por el índice point, si se sale del rango devuelve la posición del nodo
QPointF ItemNode::scenePointPos(int point) const
{
	QRectF rect = outlineRect();

	if (point < 0) return this->scenePos();
	if (point < numProp) return mapToScene( QPointF(rect.left() + lineSpacing*0.5, rect.top() + lineSpacing*(point + 1.5)) );
	if (point < 2 * numProp) return mapToScene( QPointF(rect.right() - lineSpacing*0.5, rect.top() + lineSpacing*(point-numProp + 1.5)) );
	return this->scenePos();
}

QPointF ItemNode::scenePointPos(QString name, bool in) const
{
	return scenePointPos(propPoint(name, in));
}


int ItemNode::numProps() const
{
	return numProp;
}


void ItemNode::mouseDoubleClickEvent(QGraphicsSceneMouseEvent *)
{
	window->showProperties(this);
}

// estas funciones son para conseguir que los puntos se pinten de rojo cunado se pasa sobre ellos
// y que se cree la linea al arrastar y creen los link cuando suelte en una posición correcta
void ItemNode::mousePressEvent(QGraphicsSceneMouseEvent * event)
{
	clickedPoint = pointAt(event->pos());

	if (clickedPoint < 0)
	{
		// si no coincidía con un punto mueve el item normalmente
		Node::mousePressEvent(event);
	}
	else
	{
		// si coincide con un punto crea una línea indicadora del posible enlace
		line = new QGraphicsLineItem( QLineF(event->scenePos(), event->scenePos()) );
		line->setZValue(1000);
		scene()->addItem(line);
	}
}

void ItemNode::mouseMoveEvent(QGraphicsSceneMouseEvent * event)
{
	if (clickedPoint < 0)
	{
		// si no coincidía con un punto mueve el item normalmente
		Node::mouseMoveEvent(event);
	}
	else
	{
		// desplazo la línea como el mouse, pero un poco (0.5) más corta, ya que sino cuando busquemos el item de justo
		// debajo del cursor nos puede dar la própia línea en vez del que buscamos.
		QLineF shortLine = QLineF(line->line().p1(), event->scenePos());
		shortLine.setLength(shortLine.length() - 0.5);
		line->setLine(shortLine);

		// si había un punto marcado lo desmarco
		if (externalMarkedPoint >= 0)
		{
			externalMarkedItem->unmarkPoint();
			externalMarkedPoint = -1;
		}

		// y si está sobre otro punto válido para el enlace, lo marco
		ItemNode *target = dynamic_cast<ItemNode *>(scene()->itemAt(event->scenePos())); // en un futuro tendrá que hacer el cast a Node
		if (target)
		{
			// obtengo el punto concreto a partir de las coordenadas locales al otro item
			int pointPos = target->pointAt(target->mapFromScene(event->scenePos()));
			if (pointPos >= 0)
			{ // marco el nuevo punto en función de la validez del enlace que generaría
				externalMarkedPoint = pointPos;
				externalMarkedItem = target;
				externalMarkedItem->markPoint(externalMarkedPoint, isValidLink(this, clickedPoint, target, pointPos));
			}
		}
	}
}

void ItemNode::mouseReleaseEvent(QGraphicsSceneMouseEvent * event)
{
	if (clickedPoint < 0)
	{
		// si coincide con un punto válidosi no coincidía con un punto mueve el item normalmente
		Node::mouseReleaseEvent(event);
	}
	else
	{
		// elimino la línea indicadora
		delete(line);

		// si había un punto marcado lo desmarco
		if (externalMarkedPoint >= 0)
		{
			externalMarkedItem->unmarkPoint();
			externalMarkedPoint = -1;
		}

		// si está sobre un punto válido creo el enlace
		ItemNode *target = dynamic_cast<ItemNode *>(scene()->itemAt(event->scenePos())); // en un futuro tendrá que hacer el cast a Node
		if (target)
		{
			// obtiene el punto concreto a partir de las coordenadas locales al otro item
			int pointPos = target->pointAt(target->mapFromScene(event->scenePos()));

			// si es valido crea el enlace entre clickedPoint de this y pointPos de target
			if ( (pointPos >= 0) && isValidLink(this, clickedPoint, target, pointPos) )
			{
				if (clickedPoint >= numProp) // si clickedPoint es un punto de salida, el enlace parte de this
					window->createLink(this, clickedPoint, target, pointPos);
				else // si no, parte de target
					window->createLink(target, pointPos, this, clickedPoint);
			}
		}

		clickedPoint = -1;
	}
}

void ItemNode::markPoint(int point, bool validity)
{
	markedPoint = point;
	markedValidity = validity;
	update();
}

void ItemNode::unmarkPoint()
{
	markedPoint = -1;
	update();
}

// indica si un enlace sería válido
///@todo tendría que mirar también si el tipo de las propiedades es compatible.
bool ItemNode::isValidLink(ItemNode *fromNode, int fromPoint, ItemNode *toNode, int toPoint) const
{
	int fromNumProp = fromNode->numProp;
	int toNumProp = toNode->numProp;

	if (fromNode == toNode) return false; // si son iguales
	if (fromNode->parentItem() != toNode->parentItem()) return false; // si no están al mismo nivel
	if ( (fromPoint < 0) || (toPoint < 0) ) return false; // si alguno no es un punto
	if ( (fromPoint < fromNumProp) && (toPoint < toNumProp) ) return false; // si los dos son de entrada
	if ( (fromPoint >= fromNumProp) && (toPoint >= toNumProp) ) return false; // si los dos son de salida
	if ( (fromPoint >= 2 * fromNumProp) || (toPoint >= 2 * toNumProp) ) return false; // si alguno no es un punto existente
	return true;
}

int ItemNode::insertPos() const
{
	return numProp;
}

void ItemNode::insertProperty(QString name, int type, bool input, bool output)
{
	insertProperty(insertPos(), name, type, input, output);
}

void ItemNode::insertProperty(int pos, QString name, int type, bool input, bool output)
{
	prepareGeometryChange();
	itemProp.insertProperty(pos, name, type, input, output);
	numProp = itemProp.getProperties().size();

	// recalcula el rectángulo exterior
    QFontMetricsF metrics = qApp->font();
	lineSpacing = metrics.lineSpacing();
    outlinerect = metrics.boundingRect(itemProp.getType());

	const QList<QString> props = itemProp.getProperties();
	for (int i = 0; i < props.size(); i++)
		outlinerect |= metrics.boundingRect(props[i]);

    outlinerect.adjust(0.0, 0.0, lineSpacing*4, props.size()*metrics.lineSpacing());


	// recalcula el texto a escribir
	QRectF rect = outlineRect();
	QString nameAndProp(itemProp.getType());
	for (int i = 0; i < props.size(); i++)
		nameAndProp += "\n" + props[i];

	setText(nameAndProp);
	updateLinksPos();
	update();
}

void ItemNode::removeProperty(QString name)
{
	prepareGeometryChange();
	itemProp.deleteProperty(name);
	numProp = itemProp.getProperties().size();

	// recalcula el rectángulo exterior
    QFontMetricsF metrics = qApp->font();
	lineSpacing = metrics.lineSpacing();
    outlinerect = metrics.boundingRect(itemProp.getType());

	const QList<QString> props = itemProp.getProperties();
	for (int i = 0; i < props.size(); i++)
		outlinerect |= metrics.boundingRect(props[i]);

    outlinerect.adjust(0.0, 0.0, lineSpacing*4, props.size()*metrics.lineSpacing());


	// recalcula el texto a escribir
	QRectF rect = outlineRect();
	QString nameAndProp(itemProp.getType());
	for (int i = 0; i < props.size(); i++)
		nameAndProp += "\n" + props[i];

	setText(nameAndProp);
	updateLinksPos();
	update();
}

void ItemNode::deleteProperty(int pos)
{
	prepareGeometryChange();
	itemProp.deleteProperty(pos);
	numProp = itemProp.getProperties().size();

	// recalcula el rectángulo exterior
    QFontMetricsF metrics = qApp->font();
	lineSpacing = metrics.lineSpacing();
    outlinerect = metrics.boundingRect(itemProp.getType());

	const QList<QString> props = itemProp.getProperties();
	for (int i = 0; i < props.size(); i++)
		outlinerect |= metrics.boundingRect(props[i]);

    outlinerect.adjust(0.0, 0.0, lineSpacing*4, props.size()*metrics.lineSpacing());


	// recalcula el texto a escribir
	QRectF rect = outlineRect();
	QString nameAndProp(itemProp.getType());
	for (int i = 0; i < props.size(); i++)
		nameAndProp += "\n" + props[i];

	setText(nameAndProp);
	updateLinksPos();
	update();
}
/*
void ItemNode::addNewProp(ItemNode *toNode, int toPoint)
{
	prepareGeometryChange();
	insertProperty(numProp - 1, toNode->propName(toPoint), toNode->propType(toPoint), true, false);
	update();
}
*/
void ItemNode::delLastProp()
{
	prepareGeometryChange();
	deleteProperty(numProp - 2);
	update();
}

QVariant ItemNode::itemChange(GraphicsItemChange change,
                          const QVariant &value)
{
    if (change == ItemPositionHasChanged)
		updateLinksPos();

    return QGraphicsItem::itemChange(change, value);
}

void ItemNode::updateLinksPos()
{
	foreach (Link *link, getLinks())
		link->trackNodes();
}

QRectF ItemNode::outlineRect() const
{
	return outlinerect;
}

int ItemNode::roundness(double size) const
{
    const int Diameter = 12;
    return 100 * Diameter / int(size);
}

// Devuelve el indice del punto de enlace que hay debajo de la posición "pos", en coordenadas del workernode;
// empieza a contar desde 0 y los de los de salida los cuenta a continuación de los de entrada;
// si "pos" no está sobre ningún punto de enlace válido devuelve -1.
int ItemNode::pointAt(QPointF pos) const
{
	QRectF rect = outlineRect();
	if ((pos.x() >= rect.left() + lineSpacing*0.25) && (pos.x() <= rect.left() + lineSpacing*0.75)) // si está en la columna de entrada
	{
		if (pos.y() - rect.top() >= lineSpacing*1.25) // si sestá por debajo del título
		{
			int relativePos = (int)(pos.y() - rect.top() - lineSpacing*1.25);
			if (relativePos % (int)lineSpacing <= lineSpacing*0.5) // si está sobre el lugar del punto
				if (itemProp.isInput((int)(relativePos / lineSpacing))) return (int)(relativePos / lineSpacing); // si la propiedad tiene enlace de entrada, es valido
		}
	}
	else if ((pos.x() >= rect.right() - lineSpacing*0.75) && (pos.x() <= rect.right() - lineSpacing*0.25)) // si está en la columna de salida
	{
		if (pos.y() - rect.top() >= lineSpacing*1.25) // si está por debajo del título
		{
			int relativePos = (int)(pos.y() - rect.top() - lineSpacing*1.25);
			if (relativePos % (int)lineSpacing <= lineSpacing*0.5) // si está sobre el lugar del punto
				if (itemProp.isOutput((int)(relativePos / lineSpacing))) return (int)(relativePos / lineSpacing + numProp);  // si la propiedad tiene enlace de salida, es valido
		}
	}
	return -1;
}

QString ItemNode::propName(int point) const
{
	if ( (point < 0) && (point >= 2 * numProp) ) return QString(); // si point se sale de rango

	if (point >= numProp) point = point - numProp; // ignoro si es de entrada o salida
	return itemProp.getProperties()[point];
}

int ItemNode::propPoint(QString name, bool in) const
{
	int pos = itemProp.getProperties().indexOf(name);
	if (pos < 0) return pos;
	if (!in) return pos += numProp;
	return pos;
}

int ItemNode::propType(int point) const
{
	if ( (point < 0) && (point >= 2 * numProp) ) return 0; // si point se sale de rango devuelvo "no type"

	if (point >= numProp) point = point - numProp; // ignoro si es de entrada o salida
	return itemProp.propertyType(point);
}


