/*
 *	Copyright (C) 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */


#include <QtGui>
#include <QDebug>
#include <iostream>

#include "slatewindow.h"
#include "sinclink.h"
#include "asinclink.h"
#include "sequenlink.h"
#include "node.h"
#include "groupnode.h"
#include "inputnode.h"
#include "middlenode.h"
#include "outputnode.h"
#include "slateview.h"
#include "insertaction.h"

#include <QVDesignerGUI>
//#include "../designergui.h"
#include "../facade/itemproperties.h"


SlateWindow::SlateWindow(QVDesignerGUI *desig, QWidget * parent): QMainWindow(parent), designer(desig)
{
    scene = new QGraphicsScene(-1180, -900, 3200, 2400);

    view = new SlateView;
    view->setScene(scene);
    view->setDragMode(QGraphicsView::RubberBandDrag);
    view->setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing);
    view->setContextMenuPolicy(Qt::ActionsContextMenu);
    setCentralWidget(view);

    minZ = 0;
    maxZ = 0;
    seqNumber = 0;

    createMenus();
    createToolBars();

	statusbar = new QStatusBar();
	setStatusBar(statusbar);

    connect(scene, SIGNAL(selectionChanged()), this, SLOT(updateActions()));

	scene->setBackgroundBrush(QPixmap(":/images/background1.png"));

    setWindowTitle(tr("Slate"));
    updateActions();
}

void SlateWindow::closeEvent(QCloseEvent *)
	{
	emit closed();
	}

///@todo reducir el tamaño del Pen del link en función del número de antecesores (*0.5 por cada uno)
bool SlateWindow::createLink(Node *fromNode, int fromPoint, Node *toNode, int toPoint)
{
	// si está en ejecución no se permite lincar
	if (stopAction->isEnabled()) return false;

	if (addSLinkAction->isChecked())
		designer->addLink(fromNode->getName(), fromNode->propName(fromPoint), toNode->getName(), toNode->propName(toPoint), true, false);
	else if (addALinkAction->isChecked())
		designer->addLink(fromNode->getName(), fromNode->propName(fromPoint), toNode->getName(), toNode->propName(toPoint), false, false);
	else
		designer->addLink(fromNode->getName(), fromNode->propName(fromPoint), toNode->getName(), toNode->propName(toPoint), false, true);

return true;
}

void SlateWindow::addLinkLine(QString fromNode, QString fromProp, QString toNode, QString toProp, bool sinc, bool sequential)
{
	Link *link;
	QString linkName = fromNode + "[" + fromProp + "] => " + toNode + "[" + toProp + "]";

	if ( insertNodes.contains(fromNode) && insertNodes.contains(toNode) ) {
		ItemNode *from = insertNodes.value(fromNode);
		ItemNode *to = insertNodes.value(toNode);

		if (sequential) {
			link = new SequenLink(from, fromProp, to, toProp, 0, scene);
			insertLinks.insert(linkName, link);
			link->trackNodes();
		}
		else if (sinc) {
			link = new SincLink(from, fromProp, to, toProp, 0, scene);
			insertLinks.insert(linkName, link);
			link->trackNodes();
		}
		else {
			link = new AsincLink(from, fromProp, to, toProp, 0, scene);
			insertLinks.insert(linkName, link);
			link->trackNodes();
		}
	}
}

void SlateWindow::delLinkLine(QString fromNode, QString fromProp, QString toNode, QString toProp)
{
	QString linkName = fromNode + "[" + fromProp + "] => " + toNode + "[" + toProp + "]";
	if (insertLinks.contains(linkName)) {
		delete insertLinks.value(linkName);
		insertLinks.remove(linkName);
	}
}

void SlateWindow::insertItem(QString type)
{
	QString name = type.toLower();
	name.replace('<', '_');
	name.replace(',', '_');
	name.replace('>', '_');
	name.replace(' ', "");
	name += QString("_%1").arg(seqNumber);

	designer->addItem(type, name);
}

void SlateWindow::addItemNode(QString type, QString name, ItemProperties *item)
{
	///@todo ver si los nodos al destruirse tendrían que destruir explicitamente el ItemProperties pasado
	if (designer->getInputItemTypes().contains(type)) {
		InputNode *node = new InputNode(*item, name, this, 0, scene);
		insertNodes.insert(node->getName(), node);
		setupNode(node);
		if (insertNodesPos.contains(name))
			node->setPos(insertNodesPos.value(name));
		else
			insertNodesPos.insert(name, node->pos());
	}
	else if (designer->getMiddleItemTypes().contains(type)) {
		MiddleNode *node = new MiddleNode(*item, name, this, 0, scene);
		insertNodes.insert(node->getName(), node);
		setupNode(node);
		if (insertNodesPos.contains(name))
			node->setPos(insertNodesPos.value(name));
		else
			insertNodesPos.insert(name, node->pos());
	}
	else if (designer->getOutputItemTypes().contains(type)) {
		OutputNode *node = new OutputNode(*item, name, this, 0, scene);
		insertNodes.insert(node->getName(), node);
		setupNode(node);
		if (insertNodesPos.contains(name))
			node->setPos(insertNodesPos.value(name));
		else
			insertNodesPos.insert(name, node->pos());
	}
}

void SlateWindow::delItemNode(QString name)
{
	if (insertNodes.contains(name)) {
		insertNodesPos.remove(name);
		insertNodesPos.insert(name, insertNodes.value(name)->pos());
		delete insertNodes.value(name);
		insertNodes.remove(name);
	}
}

void SlateWindow::addProperty(QString srcName, QString propName, int type, bool in, bool out)
{
	if (insertNodes.contains(srcName)) {
		insertNodes.value(srcName)->insertProperty(propName, type, in, out);
	}
}

void SlateWindow::delProperty(QString srcName, QString propName)
{
	if (insertNodes.contains(srcName)) {
		insertNodes.value(srcName)->removeProperty(propName);
	}
}

void SlateWindow::del()
{
	// primero borro solo los links, ya que al borrar un nodo pueden desaparecer links seleccionados, que se intentarian borrar y daría un core
	foreach(QGraphicsItem *item, scene->selectedItems()) {
        Link *link = dynamic_cast<Link *>(item);

        if (link && (link->fromNode() != NULL) && (link->toNode() != NULL)) {
			designer->delLink(link->fromNode()->getName(), link->fromProp(), link->toNode()->getName(), link->toProp());
        }
    }

	foreach(QGraphicsItem *item, scene->selectedItems()) {
		Node *node = dynamic_cast<Node *>(item);

		if (node) {
			designer->delItem(node->getName());
		}
    }
}

void SlateWindow::join()
{
	QList<QGraphicsItem *> group = selectedNodeGroup();

	// creo el nuevo grupo
    GroupNode *node = new GroupNode(0, scene);
    node->setText("Group");

	// si los items seleccionados estaban ya dentro de otro grupo, este se añade a el (sino a la escena)
	if ( group.first() && group.first()->parentItem() ) {
		GroupNode *parentNode = dynamic_cast<GroupNode *>(group.first()->parentItem());
		if (parentNode) parentNode->addNode(node);
	}

	// para cada item si es un nodo lo añado al nuevo grupo
	foreach (QGraphicsItem *item, group) {
		Node *itemNode = dynamic_cast<Node *>(item);
		if (itemNode) {
			node->addNode(itemNode);

			//se actualiza el tamaño de sus links
			//y si la otra punta no se va a meter al grupo pasa a formar parte de este
			foreach (Link *link, itemNode->getInLinks()) {
				Node *other = link->fromNode();
				if (group.contains(other)) {
					QPen pen = link->pen();
					pen.setWidthF(pen.width()*0.5);
					link->setPen(pen);
					link->update();
				}
/*				else {
						link->changeToNode(node);
				}*/
			}
			foreach (Link *link, itemNode->getOutLinks()) {
				Node *other = link->toNode();
				if (group.contains(other)) {
					QPen pen = link->pen();
					pen.setWidthF(pen.width()*0.5);
					link->setPen(pen);
					link->update();
				}
/*				else {
						link->changeFromNode(node);
				}*/
			}
			itemNode->updateLinksPos(); // para que se actualiza la posición y tamaño de los links del nodo
		}
	}
}

void SlateWindow::bringToFront()
{
    maxZ += 2; // dos, uno para el y otro para sus link que serán: el + 1
    setZValue(maxZ);
}

void SlateWindow::sendToBack()
{
    minZ += 2; // dos, uno para el y otro para sus link que serán: el + 1
    setZValue(minZ);
}

void SlateWindow::run()
{
	foreach (QAction *action, insertMenu->actions()) {
		action->setEnabled(false);
	}

	deleteAction->setEnabled(false);

	runAction->setEnabled(false);
	stopAction->setEnabled(true);
	designer->run();
}

void SlateWindow::stop()
{
	designer->stop();
	stopAction->setEnabled(false);
	runAction->setChecked(false);
	runAction->setEnabled(true);

    bool hasSelection = !scene->selectedItems().isEmpty();
	deleteAction->setEnabled(hasSelection);

	foreach (QAction *action, insertMenu->actions()) {
		action->setEnabled(true);
	}
}

void SlateWindow::setZValue(int z)
{
    Node *node = selectedNode();
    if (node) {
        node->setZValue(z);
		node->updateLinksPos();
	}
}

void SlateWindow::showProperties()
{
    Node *node = selectedNode();

    if (node) showProperties(node);
}

void SlateWindow::showProperties(Node *node)
{
	designer->showProperties(node->getName());
}

void SlateWindow::updateActions()
{
    bool hasSelection = !scene->selectedItems().isEmpty();
    bool isNode = (selectedNode() != 0);
//	bool isNodePair = (selectedNodePair() != NodePair());
	bool isNodeGroup = (selectedNodeGroup() != QList<QGraphicsItem *>());

	joinAction->setEnabled(isNodeGroup);
	// si está en ejecución no se permite borrar
    if (runAction->isEnabled()) deleteAction->setEnabled(hasSelection);
    bringToFrontAction->setEnabled(isNode);
    sendToBackAction->setEnabled(isNode);
    propertiesAction->setEnabled(isNode);

    foreach (QAction *action, view->actions())
        view->removeAction(action);

    foreach (QAction *action, editMenu->actions()) {
        if (action->isEnabled())
            view->addAction(action);
    }

	// añado las acciones del menu insertar al desplegable del botón derecho (de dos formas distintas)
	QAction *ma = new QAction(tr("..."), this);
	ma->setMenu(new QMenu("insert_menu", this));
	ma->menu()->addMenu(insertMenu);
	view->addAction(ma);

	foreach (QAction *action, insertMenu->actions()) {
            view->addAction(action);
    }
}

void SlateWindow::setupNode(Node *node)
{
    node->setPos( QPoint(10 + (seqNumber % 10) * 6, 20 + (seqNumber % 10) * 6) );
    ++seqNumber;

    scene->clearSelection();
    node->setSelected(true);
    bringToFront();
}

void SlateWindow::createMenus()
{
	///@todo usar el QAction.setStatusTip(string) para establecer una indicación de que hace cada acción
    exitAction = new QAction(tr("E&xit"), this);
    exitAction->setShortcut(tr("Ctrl+Q"));
    connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));

    addALinkAction = new QAction(tr("Add &Asynchronous Link"), this);
    addALinkAction->setIcon(QIcon(":/images/alink.png"));
    addALinkAction->setShortcut(tr("Ctrl+A"));
	addALinkAction->setCheckable(true); // para que se mantenga pulsada
	addALinkAction->setChecked(true); // por defecto está seleccionado asincrono

    addSLinkAction = new QAction(tr("Add &Synchronous Link"), this);
    addSLinkAction->setIcon(QIcon(":/images/slink.png"));
    addSLinkAction->setShortcut(tr("Ctrl+S"));
	addSLinkAction->setCheckable(true); // para que se mantenga pulsada

    addQLinkAction = new QAction(tr("Add Se&quential Link"), this);
    addQLinkAction->setIcon(QIcon(":/images/qlink.png"));
    addQLinkAction->setShortcut(tr("Ctrl+Q"));
	addQLinkAction->setCheckable(true); // para que se mantenga pulsada

	// con esto conseguimos meterlas en un grupo exclusivo, o una o la otra
	linkGroup = new QActionGroup(this);
	linkGroup->setExclusive(true);
	linkGroup->addAction(addALinkAction);
	linkGroup->addAction(addSLinkAction);
	linkGroup->addAction(addQLinkAction);


    joinAction = new QAction(tr("&Join"), this);
    joinAction->setIcon(QIcon(":/images/join.png"));
    joinAction->setShortcut(tr("Ctrl+J"));
    connect(joinAction, SIGNAL(triggered()), this, SLOT(join()));

    deleteAction = new QAction(tr("&Delete"), this);
    deleteAction->setIcon(QIcon(":/images/delete.png"));
    deleteAction->setShortcut(tr("Del"));
	deleteAction->setEnabled(false);
    connect(deleteAction, SIGNAL(triggered()), this, SLOT(del()));

    bringToFrontAction = new QAction(tr("Bring to &Front"), this);
    bringToFrontAction->setIcon(QIcon(":/images/bringtofront.png"));
    connect(bringToFrontAction, SIGNAL(triggered()),this, SLOT(bringToFront()));

    sendToBackAction = new QAction(tr("&Send to Back"), this);
    sendToBackAction->setIcon(QIcon(":/images/sendtoback.png"));
    connect(sendToBackAction, SIGNAL(triggered()), this, SLOT(sendToBack()));

    propertiesAction = new QAction(tr("P&roperties..."), this);
    connect(propertiesAction, SIGNAL(triggered()), this, SLOT(showProperties()));

    runAction = new QAction(tr("&Run"), this);
    runAction->setIcon(QIcon(":/images/run.png"));
	runAction->setCheckable(true); // para que se mantenga pulsada
	runAction->setChecked(true); // porque comienza en ejecución
	runAction->setEnabled(false); // porque comienza en ejecución
    connect(runAction, SIGNAL(triggered()), this, SLOT(run()));

    stopAction = new QAction(tr("&Run"), this);
    stopAction->setIcon(QIcon(":/images/stop.png"));
	stopAction->setEnabled(true); // porque comienza en ejecución
    connect(stopAction, SIGNAL(triggered()), this, SLOT(stop()));

    saveAsAction = new QAction(tr("&Export"), this);
    saveAsAction->setShortcut(tr("Ctrl+E"));
    connect(saveAsAction, SIGNAL(triggered()), this, SLOT(saveAs()));


    fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(saveAsAction);
    fileMenu->addAction(exitAction);

    editMenu = menuBar()->addMenu(tr("&Edit"));
    editMenu->addAction(addALinkAction);
    editMenu->addAction(addSLinkAction);
    editMenu->addAction(addQLinkAction);

    editMenu->addSeparator();
    editMenu->addAction(joinAction);
    editMenu->addAction(deleteAction);
    editMenu->addSeparator();
    editMenu->addAction(bringToFrontAction);
    editMenu->addAction(sendToBackAction);
    editMenu->addSeparator();
    editMenu->addAction(propertiesAction);


	// generamos el insert menu a partir de la lista que nos devuelve el controlador
	insertMenu = menuBar()->addMenu(tr("&Insert"));

	foreach (QString item, designer->getItemTypes())
		{
		InsertAction *action = new InsertAction(item, this);
		action->setEnabled(false);
		connect(action, SIGNAL(triggered(QString)), this, SLOT(insertItem(QString)));
		insertMenu->addAction(action);
		}
}

void SlateWindow::includeItemType(QString itemType)
{
	InsertAction *action = new InsertAction(itemType, this);
	action->setEnabled(false);
	connect(action, SIGNAL(triggered(QString)), this, SLOT(insertItem(QString)));
	insertMenu->addAction(action);

	updateActions();
}

void SlateWindow::createToolBars()
{
    editToolBar = addToolBar(tr("Edit"));
    editToolBar->addAction(addALinkAction);
    editToolBar->addAction(addSLinkAction);
    editToolBar->addAction(addQLinkAction);
    editToolBar->addAction(joinAction);
    editToolBar->addAction(deleteAction);
    editToolBar->addSeparator();
    editToolBar->addSeparator();
    editToolBar->addAction(bringToFrontAction);
    editToolBar->addAction(sendToBackAction);

	editToolBar->addSeparator();
	editToolBar->addAction(runAction);
	editToolBar->addAction(stopAction);
}

Node *SlateWindow::selectedNode() const
{
    QList<QGraphicsItem *> items = scene->selectedItems();
    if (items.count() == 1) {
        return dynamic_cast<Node *>(items.first());
    } else {
        return 0;
    }
}

Link *SlateWindow::selectedLink() const
{
    QList<QGraphicsItem *> items = scene->selectedItems();
    if (items.count() == 1) {
        return dynamic_cast<Link *>(items.first());
    } else {
        return 0;
    }
}

SlateWindow::NodePair SlateWindow::selectedNodePair() const
{
    QList<QGraphicsItem *> items = scene->selectedItems();
    if (items.count() == 2) {
        Node *first = dynamic_cast<Node *>(items.first());
        Node *second = dynamic_cast<Node *>(items.last());
        if (first && second)
            return NodePair(first, second);
    }
    return NodePair();
}

// devuelve la lista de onlyParents seleccionados, si son almenos 2 comparten padre (tambien devuelve los link de dicha lista)
QList<QGraphicsItem *> SlateWindow::selectedNodeGroup() const
{
    QList<QGraphicsItem *> items = onlyParents(scene->selectedItems());

	//si quedan almemos dos
    if (items.count() > 1) {
		// y si comparten padre
		QGraphicsItem * parent = 0;
		QMutableListIterator<QGraphicsItem *> i(items);
		if (i.hasNext()) parent = i.next()->parentItem();

		while (i.hasNext())
			if ( i.next()->parentItem() != parent) return QList<QGraphicsItem *>();

		return items;
    }
    return QList<QGraphicsItem *>();
}

// obtiene los QGraphicsItem de la lista cuyo padre no pertenece a la lista
QList<QGraphicsItem *> SlateWindow::onlyParents(QList<QGraphicsItem *> items) const
{
    QList<QGraphicsItem *> parentItems;

    QMutableListIterator<QGraphicsItem *> i(items);
    while (i.hasNext()) {
		i.next();
		bool parent = true;
		QMutableListIterator<QGraphicsItem *> j(items);
		while (j.hasNext()) {
			j.next();
			if ( (i.value()) && (i.value()->parentItem() == j.value()) ) {
				parent = false;
				break;
			}
		}
		if (parent) parentItems.append(i.value());
    }

	return parentItems;
}

bool SlateWindow::saveAs()
{
    QString fileName = QFileDialog::getSaveFileName(this,
                               tr("Save C++"), ".",
                               tr("C++ files (*.cpp)"));
    if (fileName.isEmpty())
        return false;
	return saveFile(fileName);
}

bool SlateWindow::saveFile(const QString &fileName)
{
    if (!writeCppFile(fileName)) {
//         statusBar()->showMessage(tr("Saving canceled"), 2000);
        return false;
    }
//     setCurrentFile(fileName);
//     statusBar()->showMessage(tr("File saved"), 2000);
    return true;
}

bool SlateWindow::writeCppFile(const QString &fileName)
{
    QFile file(fileName);
    if (!file.open(QIODevice::WriteOnly)) {
        QMessageBox::warning(this, tr("QVDesignerGUI"),
                             tr("Cannot write file %1:\n%2.")
                             .arg(file.fileName())
                             .arg(file.errorString()));
        return false;
    }

    QTextStream out(&file);
    out << designer->getCppText();

    return true;
}


void SlateWindow::showMessage(QString message)
{
	if (statusbar)
		statusbar->showMessage(message, 10000); // lo muestra 10 segundos
}


void SlateWindow::arrangeItems()
{
	QList<QList<ItemNode *> *> levels;
	int left = 10;
	int bottom = 20;
	const int MARGIN = 20;

	// obtengo los niveles en que se encuentran los items, organizandolos por columnas en "levels"
	foreach(ItemNode *node, insertNodes) {
		int level = node->precursors();
		while(levels.size() <= level)
			levels.append(new QList<ItemNode *>);
		levels[level]->append(node);
	}

	//muevo los items en la pizarra en función de los tamaños y niveles de estos
	foreach(QList<ItemNode *> *levelNodes, levels) {
		int maxWidth = 0;
		if (levelNodes) {
			foreach(ItemNode *node, *levelNodes) {
				node->setPos(QPoint(left, bottom));
				bottom += (int)node->boundingRect().height() + MARGIN; //le sumo a bottom el alto mas un margen
				int width = (int)node->boundingRect().width();
				if (maxWidth < width) maxWidth = width; //actualiza maxWidth
			}
		}
		left += maxWidth + MARGIN; //le sumo a left maxWide mas un margen
		bottom = 20;//reseteo bottom
	}

	// vacio la estructura usada
	foreach(QList<ItemNode *> *levelNodes, levels) {
		levelNodes->clear();
		delete levelNodes;
	}
}


