/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmplayercamera.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVMPLAYERCAMERA_H
#define QVMPLAYERCAMERA_H

#include <QVMPlayerProxy>

/*!
@class QVMPlayerCamera qvmplayerproxy.h qvio/qvmplayerproxy.h
@brief Class to create versatile video input blocks for the QVision application architecture.

This is an implementation of the class @ref QVCamera, based on the @ref QVMPlayerProxy interface class for the <a href="http://www.mplayerhq.hu/">MPlayer</a> application.

This class can create input blocks to read frames from a video source, in a QVision application. The video sources can be any video file or device connected physically to the computer, or remotely through a network video stream. The class can read from a wide variety of media formats, becoming is a good, robust and versatile video input reader class.

For a detailed usage of this class in a QVision application, see section @ref ProgrammingModel. To learn how to specify a given video source for the class, read @ref QVMPlayerProxy class documentation. Specifically, the @ref QVMPlayerURLFormat section.

@see QVMPlayerProxy
@ingroup qvio
*/
class QVMPlayerCamera : public QVCamera
	{
	Q_OBJECT
	public:
		/// @brief Constructs a QVMPlayerCamera, which is initially in "Closed" status.
		QVMPlayerCamera(QString name = QString());
		/// @brief Destroys a QVMPlayerCamera.
		~QVMPlayerCamera();

		#ifndef DOXYGEN_IGNORE_THIS
		/// @brief Opens a video source for image reading.
		/// This is an overloaded version for @ref openCam(const QString &, unsigned int, unsigned int, OpenOptions)
		/// function that will get the url and the configuration options from dynamic properties defined in the camera
		/// object.
		///
		/// Used properties:
		///	 - <b>URL</b> (QString)		<br>URL identifying the device and opening options to open.
		///					See @ref QVMPlayerURLFormat for URL formats.
		///	 - <b>Rows</b> (int)		<br>Number of requested output rows.
		///	 - <b>Cols</b> (int)		<br>Number of requested output cols.
		///	 - <b>RealTime</b> (bool)	<br>If the camera should be opened in RealTime mode
		///	 - <b>Deinterlaced</b> (bool)	<br>If the camera should be opened in Deinterlaced mode
		///	 - <b>RGBMEncoder</b> (bool)	<br>If the camera should be opened in RGBMEncoder mode
		///	 - <b>NoLoop</b> (bool)		<br>If the camera should be opened in NoLoop mode
		///
		/// For further info about the usage of these parameters see QVMPlayerCamera::OpenOption flags.
		/// @return True if the camera was opened successfully, false otherwise.
		/// @see openCam(const QString &, unsigned int, unsigned int, OpenOptions)
		bool openCam();
		#endif

		/// @brief Returns the current camera name (without path).
		/// @returns Current camera name.
		/// @deprecated Use getName() instead
		const QString getUrlBase() const { return mplayerProxy.getUrlBase(); }

		/// @brief Returns current camera options.
		/// @returns Current camera options.
		QVMPlayerProxy::OpenOptions getOptions() const		{ return mplayerProxy.getOptions(); };

		/// @brief Returns the total number of updated frames.
		/// @returns Current total number of updated frames.
		unsigned int getFramesUpdated() const			{ return mplayerProxy.getFramesUpdated(); }

		/// @brief Returns the number of rows of the buffer.
		/// @returns Number of rows.
		unsigned int getRows() const				{ return mplayerProxy.getRows(); };

		/// @brief Returns the number of cols of the buffer.
		/// @returns Number of cols.
		unsigned int getCols() const				{ return mplayerProxy.getCols(); };

		/// @brief Returns the camera speed in frames per second.
		/// @returns Camera speed in frames per second.
		double getFPS() const					{ return mplayerProxy.getFPS(); };

		/// @brief Returns the video file length in seconds.
		/// @returns Video length in seconds.
		double getTimeLength() const				{ return mplayerProxy.getTimeLength(); };

		/// @brief Returns the current video position in seconds.
		/// @returns Current video position in seconds.
		double getTimePos() const				{ return mplayerProxy.getTimePos(); };

		/// @brief Returns the current video speed.
		/// @returns Current video speed (as a multiplier of the FPS).
		double getSpeed() const					{ return mplayerProxy.getSpeed(); };

		/// @brief Returns number of frames grabbed.
		/// @returns Number of successful frames returned by grab's
		unsigned int getFramesGrabbed() const			{ return mplayerProxy.getFramesGrabbed(); };

		/// @brief Returns number of frames read.
		/// @returns Number of frames actually read by a camera.
		unsigned int getFramesRead() const			{ return mplayerProxy.getFramesRead(); };

		/// @brief Tells if the camera is a live camera.
		/// @returns True if camera is a live camera.
		bool isLiveCamera() const				{ return mplayerProxy.isLiveCamera(); };

	public slots:
		/// @brief Grabs a frame from camera input.
		///
		/// Grabs an image, and stores it in the corresponding properties of the camera:
		///	"RGB image", "Y channel image", "U channel image", and "V channel image".
		/// @param image QVImage that will store grabbed frame.
		/// @returns TRUE if success, FALSE if fail.
		bool grab();

		/// @brief Pauses the camera.
		void pauseCam()						{ mplayerProxy.pauseCam(); }

		/// @brief Unpauses the camera.
		void unpauseCam()					{ mplayerProxy.unpauseCam(); }

		/// @brief Go to the next frame.
		void nextFrameCam()					{ mplayerProxy.nextFrameCam(); }

		/// @brief Set the camera speed.
		/// @param d Camera speed (as a multiplier of the FPS)
		void setSpeedCam(double d)				{ mplayerProxy.setSpeedCam(d); }

		/// @brief Moves the camera file to a given position.
		/// @param type Type of seek: Current=0,Percentage=1,Absolute=2.
		/// @param pos 
		///	- If in \c Current mode: Relative displacement with respect to the current position in seconds.
		///	- If in \c Percentage mode: Percentage of total video length.
		///	- If in \c Absolute mode: Absolute position in seconds.
		void seekCam(QVCamera::TSeekType type, double pos)	{ mplayerProxy.seekCam(type, pos); }

		/// @brief Closes the camera.
		void closeCam();

		bool linkProperty(QString sourcePropertyName, QVPropertyContainer *destinyContainer, QString destinyPropertyName, LinkType linkType = AsynchronousLink);
		bool linkProperty(QString sourcePropertyName, QVPropertyContainer &destinyContainer, QString destinyPropertyName, LinkType linkType = AsynchronousLink);
		bool linkProperty(QVPropertyContainer *destinyContainer, QString destinyPropertyName, LinkType linkType = AsynchronousLink);
		bool linkProperty(QVPropertyContainer &destinyContainer, QString destinyPropertyName, LinkType linkType = AsynchronousLink);
		bool linkProperty(QString sourcePropertyName, QVPropertyContainer *destinyContainer, LinkType linkType = AsynchronousLink);
		bool linkProperty(QString sourcePropertyName, QVPropertyContainer &destinyContainer, LinkType linkType = AsynchronousLink);
		void linkProperty(QVPropertyContainer *container, LinkType linkType = AsynchronousLink);
		void linkProperty(QVPropertyContainer &container, LinkType linkType = AsynchronousLink);
		bool unlinkProperty(QString origName, QVPropertyContainer *destCont, QString destName);
		bool unlinkProperty(QString origName, QVPropertyContainer &destCont, QString destName);
		void unlink();

		bool link(QVWorker *worker, const QString imageY, const QString imageU, const QString imageV);
		bool link(QVWorker &worker, const QString imageY, const QString imageU, const QString imageV);
		bool unlink(QVWorker *worker, const QString imageName);
		bool unlink(QVWorker &worker, const QString imageName);

	private:
		// this boolean value becomes true if any RGB image is linked to the camera, from a worker.
		bool rgbMode;
		QVMPlayerProxy mplayerProxy;

		/// @todo ERROR: descomentar esto produce core.
		QStringList mplayer_args;		// MPlayer arguments
		QString path, schema;			// path and schema extracted from the URL
	};

#endif




