/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmplayerreader.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVMPLAYERREADER_H
#define QVMPLAYERREADER_H

#include <QFile>
#include <QProcess>
#include <QString>
#include <QThread>
#include <QUrl>

#include <qvutils/qnamedpipe.h>
#include <QVImage>

#ifndef DOXYGEN_IGNORE_THIS

// Auxiliary QVCheckOKMPlayer Class:
// This is an internal convenience thread used only to unblock the open function
// when it is waiting for mplayer to write in the fifo, in the case when
// mplayer aborted when starting (due to bad URL, or wrong file format, for
// example).
class QVCheckOKMPlayer: public QThread
{
	friend class QVMPlayerReader; // Only QVMPlayerReader will have access to this class.

	Q_OBJECT
	private:
		QVCheckOKMPlayer(QFile & fifo_file,int max_time_ms_to_wait_for_open);

		QFile & _fifo_file;
		int _max_time_ms_to_wait_for_open;

		void run();

	private slots:
		void writeErrorInFifo();
};

#endif

/*!
@class QVMPlayerReader qvmplayerreader.h qvio/qvmplayerreader.h
@brief Class to read from many video formats and devices, based on the MPlayer application.

This class is based on the <a href="http://www.mplayerhq.hu/">MPlayer</a> application. It can read videos from a wide variety of media formats, webcams and digital cameras.

Internally, when used to read from a video source compatible with MPlayer, the @ref QVMPlayerReader object will launch the MPlayer application to open and read from a video source. MPlayer will be configured to redirect its output to a system pipeline (named fifo), from which the @ref QVMPlayerRead class will read video frames.

You will probably never have to directly use this class. Usually, you will directly use a @ref QVMPlayerCameraWorker instead, which internally (and transparently) uses this class. Anyway, it is important for you to know well the URL syntax for especifying a video source:

\section QVMPlayerURLFormat Video source identifier URL formats
Each video source is identified by a URL string, which will denote video files, webcams, remote videos, etc...

\subsection URLSimplifiedFormat Simplified camera URL format.
Format: just the path to a file or files. Type of formats
 - For video files: 
	\verbatim path-to-file/file-name.{avi, mpg, dv, ...} \endverbatim
 - For devices (webcams, digital cams, tv inputs):
	\verbatim /dev/devicename \endverbatim
 - For image lists:
	\verbatim path-to-directory/filemask.{jpg, png, ...} \endverbatim

\subsection ExtendedFormat Extended camera URL formats.
The previous formats are the mor common ones, but they can be insufficient in many cases. The more general URL formats are the following (depending on the type of the video source):

Notation: [] means optional, {} means alternative:
	- Local video file compatible with mplayer:
		\verbatim file://[path-to-file/]file-name.{avi, mpg, dv, ...} \endverbatim
		Path can be relative, or absolute. Absolute paths begin with an extra slash. Examples:
		\verbatim file:///home/vision/penguin.dv \endverbatim
		\verbatim file://../../home/vision/penguin.dv \endverbatim
	- Remote video file (via http, rtsp, ftp):
		\verbatim {http,rtsp,ftp}://domain-name/path-to-file/file.{avi, mpg, dv, ...} \endverbatim
		Examples:
		\verbatim rtsp://ficheros.7rm.es/7RM/7rm06-01-071430.rm \endverbatim
		\verbatim http://perception.inf.um.es/public_data/videos/misc/penguin.dv \endverbatim
		\verbatim ftp://user:passwd@server.dom/data/video.avi \endverbatim

	- DVD | VCD video track (see mplayer man page for correct URL 
		formation for a specific title, track, device, etc.).		
		\verbatim {dvd,vcd}://{video track}\endverbatim
		Examples:
		\verbatim dvd://2 \endverbatim
		\verbatim vcd://1 \endverbatim

	- Sequence of multiple JPEG or PNG files:
		\verbatim mf://path-to/filemask.{jpg, png, ...} \endverbatim
		Examples:
		\verbatim mf:///home/user/images/img*.jpg \endverbatim
		\verbatim mf:///./images/img*.png \endverbatim

	- Video 4 linux, video 4 linux 2 camera, and analog inputs:
		\verbatim {v4l,v4l2,analog}://{device file}[?options] \endverbatim
		Examples:
		\verbatim v4l:///dev/video0 \endverbatim
		\verbatim analog:///dev/video0 \endverbatim
		See section @ref CameraParameters por options.

\subsection CameraParameters Camera parameters
Both extended and simplified formats can include special parameters for tuning device input options. Some of these parameters are:
	- <b>width=&lt;value&gt;</b>	<br>sets the width of the image frames to be read from the video device.
	- <b>height=&lt;value&gt;</b>	<br>sets the height of the image frames to be read from the video device.
	- <b>fps=&lt;value&gt;</b>	<br>sets the number of frames per second to be read from the video device.
	- <b>[brightness|contrast|hue|saturation]=[-100, 100]</b>
					<br>sets the brightness, contrast, hue or saturation for the device.
	- <b>channel=&lt;value&gt;</b>	<br>specifies the tv channel to read images from.
You can get a complete list of these options in the MPlayer documentation (option <i>-tv</i>).

Setting the width and height for the images read from the input device using parameters <i>width</i> and <i>height</i> differs from using r and c parameters of the openCam() method. The former can configure the device to capture images using a higher or lesser resolution, while the latter simply (post-)scale the images as MPlayer reads them from the input device in their original sizes.

These options work for video 4 Linux, video 4 Linux 2 camera, and analog inputs. They can be included appending a <i>?</i> sign after the URL for the device, and separated with <i>&</i> signs. Some example usages of these URL options follow:
	- URL to open a V4L2 device specifying 25 frames per second, image size of 640x480, and minimum brightness:
		\verbatim v4l2:///dev/video0?height=480&width=640&brightness=-100&fps=25 \endverbatim
	- The same device and configuration using a simplified URL:
		\verbatim /dev/video0?height=480&width=640&brightness=-100&fps=25 \endverbatim
	- URL to open an tv device at a 320x240 resolution, on channel 42
		\verbatim analog:///dev/tv?height=240&width=320&channel=42 \endverbatim

@see QVMPlayerCameraWorker
@ingroup qvio
*/

class QVMPlayerReader : public QObject
	{
	Q_OBJECT
	public:
		/// @brief Constructs a QVMPlayerReader.
		QVMPlayerReader(): open_options(Default), path(QString()), schema(QString()), camera_opened(FALSE), frames_grabbed(0), live_camera(FALSE), imgY(QVImage<uChar>()), imgU(QVImage<uChar>()), imgV(QVImage<uChar>()), cols(0), rows(0), fps(0),  time_length(0), time_pos(0), end_of_video(FALSE) { };

		/// @brief Destroys a QVMPlayerReader.
		~QVMPlayerReader() { if (camera_opened) closeCam(); };

		/// @brief Open options for a mplayer camera. Combine them using OR (|).
		enum OpenOption {
			/// Default options: don't include any of the other options.
			Default = 0x0,
			/// Mplayer will NOT ignore frame duration (i.e., will not use -benchmark, as default).
			RealTime = 0x1,
			/// Make mplayer deinterlace the image (odd lines).
			Deinterlaced = 0x2,
			/// Continue from the beginning of the video when it reaches the end.
			NoLoop = 0x4
		};

		/// @brief Options for a seek through the file.
		typedef enum { SeekCurrent = 0, SeekPercentage = 1, SeekAbsolute = 2 } TSeekType;

		Q_DECLARE_FLAGS(OpenOptions,OpenOption);

		/// @brief Opens a mplayer source given by its URL, in a requested mode and output scale size.
		///
		/// \b Note: Please observe that mplayer could slightly change the suggested size when playing
		///  the video (for example, an even number of cols is not allowed, due to YUV422 format restrictions).
		///  You can use methods getCols() and getRows() after opening to confirm the correct values.
		///
		/// @param urlstring Camera identifying URL string. See @ref QVMPlayerURLFormat for URL formats.
		/// @param opts Camera opening options (see QVMPlayerReader::openOption)
		/// @param suggested_cols Number of requested output cols (0 => default width of video).
		/// @param suggested_rows Number of requested output rows (0 => default height of video).
		/// @return True if the camera was opened successfully, false otherwise.
		bool openCam(const QString & urlstring, OpenOptions opts = Default, unsigned int suggested_cols = 0, unsigned int suggested_rows = 0);

		/// @brief Grabs a grayscale frame.
		///
		/// @param image QVImage<uChar,1> that will store grabbed frame.
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar,1> &image);

		/// @brief Grabs a RGB frame.
		///
		/// @param imageRGB QVImage<uChar,3> that will store grabbed frame.
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar,3> & imageRGB);

		/// @brief Grabs a frame in YUV format.
		///
		/// This function grabs the image in YUV format, storing each channel Y, U and V, in a separate QVImage of
		/// one channel (note that YUV format is 422, thus QVImage corresponding to plane Y will be four times larger
		/// than those corresponding to planes U and V).
		///
		/// @param imgY QVImage<uChar,1> that will store grabbed frame Y channel.
		/// @param imgU QVImage<uChar,1> that will store grabbed frame U channel (half sized cols and rows).
		/// @param imgV QVImage<uChar,1> that will store grabbed frame V channel (half sized cols and rows).
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV);

	public slots:
		/// @brief Closes the camera.
		void closeCam();

		/// @brief Moves the camera file to a given position  (if applicable; not valid for live cameras, for example).
		/// @param type Type of seek: SeekCurrent=0, SeekPercentage=1, SeekAbsolute=2.
		/// @param pos 
		///	- If in \c Current mode: Relative displacement with respect to the current position in seconds.
		///	- If in \c Percentage mode: Percentage of total video length.
		///	- If in \c Absolute mode: Absolute position in seconds.
		void seekCam(TSeekType type, double pos);

	signals:
		void camOpened();
		void camClosed();
		void newGrab();

	public:
		/// @brief Returns the current camera name (without path).
		/// @returns Current camera name.
		/// @deprecated Use getName() instead
		const QString getUrlBase() const		{ return path.split("/").last(); }

		/// @brief Returns current camera options.
		/// @returns Current camera options.
		OpenOptions getOptions() const			{ return open_options; };

		/// @brief Returns the number of cols of the buffer.
		/// @returns Number of cols.
		unsigned int getCols() const			{ return cols; };

		/// @brief Returns the number of rows of the buffer.
		/// @returns Number of rows.
		unsigned int getRows() const			{ return rows; };

		/// @brief Returns the camera speed in frames per second.
		/// @returns Camera speed in frames per second.
		double getFPS() const				{ return fps; };

		/// @brief Returns the video file length in seconds.
		///
		/// \b Note: Please observe that mplayer does not tell us the length of the video in seconds until it
		/// has read a few frames (commonly two or three). Thus, this function will return 0.0 until a few calls to
		/// any grab() function have been made.
		///
		/// @returns Video length in seconds.
		double getTimeLength() const			{ return time_length; };

		/// @brief Returns the current video position in seconds.
		/// @returns Current video position in seconds.
		double getTimePos() const			{ return time_pos; };

		/// @brief Returns number of frames grabbed.
		/// @returns Total number of frames grabbed by the camera.
		unsigned int getFramesGrabbed() const		{ return frames_grabbed; };

		/// @brief Tells if the camera is a live camera.
		/// @returns True if camera is a live camera (i.e., a webcam, a TV, etc., but not a file, for example).
		bool isLiveCamera() const			{ return live_camera; };

	private:
		OpenOptions open_options;
		QStringList mplayer_args;		// MPlayer arguments
		QProcess *mplayer;			// Mplayer process
		QString path, schema;			// Path and schema extracted from the URL
		QNamedPipe *namedPipe;			// The named fifo
		QFile fifoInput;			// The fifo file
		bool camera_opened;
		unsigned int frames_grabbed;
		bool live_camera;
		QVImage<uChar> imgY, imgU, imgV;
		int cols, rows, fps;
		double time_length, time_pos;
		bool end_of_video;

		void initMPlayerArgs(QString urlString, unsigned int suggested_cols, unsigned int suggested_rows);
		int interpretMPlayerOutput();
		bool performGrab();

	};

// Camera opening flags | operator:
Q_DECLARE_OPERATORS_FOR_FLAGS(QVMPlayerReader::OpenOptions)

#endif
