/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvyuv4mpeg2cameraworker.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVYUV4MPEG2CAMERAWORKER_H
#define QVYUV4MPEG2CAMERAWORKER_H

#include <QTimer>

#include <QVWorker>
#include <qvio.h>

#include <qvio/qvcameraworker.h>

/*!
@class QVYUV4MPEG2CameraWorker qvyuv4mpeg2cameraworker.h qvio/qvyuv4mpeg2cameraworker.h
@brief Class to create yuv4mpeg2 video file reader "input blocks" for the QVision application architecture.

This is an implementation of the class @ref QVCameraWorker, which can be used to create video input blocks for a QVision application architecture. These blocks can read image frames from video files stored in the yuv4mpeg2 format.

To use it, create a QVYUV4MPEG2CameraWorker object in the <i>main function</i>, and link a gray image, a RGB image, or three (Y,U,V) one channel images from a worker. For example:

@code
int main(int argc, char *argv[])
	{
	[...]
	// The following worker should have a property named 'Output image' containing a gray (QVImage<uChar>) or 
	// (QVImage<uChar,3>) RGB image:
	MyWorker myWorker("My worker");
	[...]
	QVYUV4MPEG2CameraWorker camera("Input camera");
	camera.linkUnspecifiedOutputProperty(&myWorker, "Input image");
	[...]
	}
@endcode

Alternatively you can link simultaneously three properties containing the Y, U and V channels of a YUV image respectively to store  images in the video file directly in that format:

@code
int main(int argc, char *argv[])
	{
	[...]
	// The following worker should have three QVImage<uChar,1> properties named 'Input image Y', 'Input image U',
	// and 'Input image V':
	MyWorker myWorker("My worker");
	[...]
	QVYUV4MPEG2CameraWorker camera("Input camera");
	camera.linkUnspecifiedOutputProperty(&myWorker, "Input image Y", "Input image U", "Input image V");
	[...]
	}
@endcode

Input properties are the following:

- <i>NoLoop</i> (bool): If the YUV4MPEG2 camera should be opened in no loop mode.
- <i>RealTime</i> (bool): If the YUV4MPEG2 camera should be opened in real time mode.
- <i>URL</i> (QString): Name of the video file to read.
- <i>Cols</i> (int): Number of suggested columns for the video source.
- <i>Rows</i> (int): Number of suggested rows for the video source.

<b>Note</b>.- RealTime property is also input (and as such can be initialized using the command line), but invisible to the GUI (because once the camera has been linked, it cannot be relinked changing its synchronism).

<b>Note</b>.- Note that you can open the camera in a different size of that specified in the real video source file (this class will automatically rescale output images if needed).

Output properties are the following:

- <i>Opened</i> (bool): Tells if the camera is correctly opened and working.
- <i>FPS</i> (int): FPS of the video.
- <i>Frames</i> (int): Number of read frames up to now -not necessarily equal to the number of processed frames, if in real time mode-.
- <i>ColsR</i> (int): Actual number of columns of the video -not necessarily equal to the number of requested columns, which will in any case be transparently rounded to the nearest upper even integer-.
- <i>RowsR</i> (int): Actual number of rows of the video -not necessarily equal to the number of requested rows, which will in any case be transparently rounded to the nearest upper even integer-..

<b>Note</b>.- Remember that you can open the camera in a different size of that specified in the real video source file (this class will automatically rescale output images if needed); thus, <i>ColsR</i> and <i>RowsR</i> will always contain the real size of the output images served by this worker.

Output image properties:

- <i>RGB image</i> (QVImage<uChar,3>): Last grabbed RGB image.
- <i>Y channel image</i> (QVImage<uChar,1>): Last grabbed Y channel image.
- <i>U channel image</i> (QVImage<uChar,1>): Last grabbed U channel image.
- <i>V channel image</i> (QVImage<uChar,1>): Last grabbed V channel image.

You can specify any of the input properties at the start of the program, using the command line parameters; for example, the file to read the video from can be specified by calling the program this way: ./program --File="/home/me/myfile.yuv". But modification of these properties can also be done directly in the corresponding graphical widgets when reopening the camera during execution (though, of course, these new values will have no effect until the camera is reopened).

Two important properties are <i>RealTime</i> and <i>NoLoop</i>, both boolean. The first one specifies if the camera will be read in real time mode -possibly leaving frames unprocessed, if processing is slower than reading- (RealTime=true), or at maximum speed -but without leaving any frame unprocessed in any case- (RealTime=false). The second one specifies if the camera will be closed when the end of the file is reached (NoLoop=true), or will loop, starting from the first frame again and again (NoLoop=false).

Output image properties can be linked automatically without the need of naming them explicitly, using the @ref QVCameraWorker::linkUnspecifiedOutputProperty() method.

@ingroup qvio
*/

class QVYUV4MPEG2CameraWorker: public QVCameraWorker
	{
	public:
		/// @brief Constructor for QVYUV4MPEG2CameraWorker class.
		///
		/// @param name the (optional) name given to the camera worker.
		QVYUV4MPEG2CameraWorker(QString name = QString("")) : QVCameraWorker(name) { };

		/// @brief Destructor for QVYUV4MPEG2CameraWorker class.
		~QVYUV4MPEG2CameraWorker() { closeCam(); };

	private:
		QFile videoFile;
		unsigned int inp_cols,inp_rows,out_cols,out_rows;
		unsigned int out_fps;
		bool rescale;
		bool realTime;

	protected:
		bool openCam(QString urlName, int &cols, int &rows, int &fps);
		void closeCam();
		bool grab(QVImage<uChar,1> &imgY, QVImage<uChar,1> &imgU, QVImage<uChar,1> &imgV);
	};

#endif
