/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvip.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#include <qvip.h>
#include <qvipp.h>

void FilterLocalMax(const QVImage<sFloat> &src, QVImage<uChar> &dest, uInt colMaskSize, uInt rowMaskSize, sFloat threshold)
	{
	const int cols = src.getCols(), rows = src.getRows();
	Set(0, dest);
	sFloat actual;
	QVIMAGE_INIT_READ(sFloat,src);
	QVIMAGE_INIT_WRITE(uChar,dest);
	for(int row = rowMaskSize; row < rows-rowMaskSize; row++)
		for(int col = colMaskSize; col < cols-colMaskSize; col++)
			{
			actual = QVIMAGE_PIXEL(src, col, row,0);
			if (actual >= threshold)
				{
				QVIMAGE_PIXEL(dest, col, row, 0) = IPP_MAX_8U;
				for (int j = row-rowMaskSize; (j < row+rowMaskSize) && (QVIMAGE_PIXEL(dest, col, row, 0) > 0); j++)
					for (int i = col-colMaskSize; i < col+colMaskSize; i++)
						if ( ((i != col) || (j != row)) && (actual <= QVIMAGE_PIXEL(src, i, j, 0)) )
							{
							QVIMAGE_PIXEL(dest, col, row, 0) = 0;
							break;
							}
				}
			}
	}

void FilterHarrisCornerResponseImage(const QVImage<uChar> &image, QVImage<sFloat> &result, const QPoint &destROIOffset)
	{
	const uInt rows = image.getRows(), cols = image.getCols();
	QVImage<uChar> buffer;
	MinEigenValGetBufferSize(image, buffer);
	QVImage<sFloat>	eigval(cols, rows);

	MinEigenVal(image, eigval, ippKernelSobel, 3, 5, buffer);

	result = QVImage<sFloat>(cols, rows);

	FilterEqualizeHistogram(eigval, result, destROIOffset);
	}

void FilterDoG(const QVImage<uChar> &image, QVImage<sFloat> &result)
	{
	const uInt rows = image.getRows(), cols = image.getCols();
	QVImage<uChar> gauss3x3(cols, rows), gauss5x5(cols, rows);
	
	FilterGauss(image, gauss3x3, ippMskSize3x3);
	FilterGauss(image, gauss5x5, ippMskSize5x5);

	result = QVImage<sFloat>(cols, rows);
	AbsDiff(gauss3x3, gauss5x5, result);	
	}

void SobelCornerResponseImage(const QVImage<sFloat> &image, QVImage<sFloat> &result)
	{
	std::cerr << "WARNING: SobelCornerResponseImage is deprecated. Use FilterHessianCornerResponseImage instead." << std::endl;
	FilterHessianCornerResponseImage(image, result);
	}

void FilterHessianCornerResponseImage(const QVImage<sFloat> &image, QVImage<sFloat> &result, const QPoint &destROIOffset)
	{
	const uInt	cols = image.getCols(), rows = image.getRows();
	//const QVImage<sFloat> imageSFloat = image;

	QVImage<sFloat> Gx(cols, rows), Gy(cols, rows), Gxx(cols, rows),
			Gxy(cols, rows), Gyx(cols, rows), Gyy(cols, rows);

	FilterSobelHorizMask(image,Gx);
	FilterSobelVertMask(image,Gy);

	FilterSobelHorizMask(Gx,Gxx);
	FilterSobelVertMask(Gy,Gyy);

	FilterSobelHorizMask(Gy,Gyx);
	FilterSobelVertMask(Gx,Gxy);

	QVImage<sFloat> GxyGyx(cols, rows), GxxGyy(cols, rows);

	Mul(Gxy, Gyx, GxyGyx);
	Mul(Gxx, Gyy, GxxGyy);

	QVImage<sFloat> diffGxyGyx_GxxGyy(cols, rows);

	Sub(GxyGyx, GxxGyy, diffGxyGyx_GxxGyy);

	QVIMAGE_INIT_WRITE(sFloat,diffGxyGyx_GxxGyy);
	for (uInt col = 0; col < cols; col++)
		for(uInt row = 0; row < rows; row++)
			if (QVIMAGE_PIXEL(diffGxyGyx_GxxGyy, col, row, 0) <= 0)
				QVIMAGE_PIXEL(diffGxyGyx_GxxGyy, col, row, 0) = 0;

	result = QVImage<sFloat>(cols, rows);

	FilterEqualizeHistogram(diffGxyGyx_GxxGyy, result, destROIOffset);
	}


int myFloodFill(QVImage<uChar> &image, uInt x, uInt y, uInt value, uInt minVal, uInt maxVal)
	{
	// Value should be inside range [minVal, maxVal]
	Q_ASSERT( (value <= minVal) || (value >= maxVal) );
	Q_ASSERT( minVal <= maxVal );

	// Coordinates should be inside the image.
	if ( (x >= image.getCols()) || (y >= image.getRows()))
		return 0;

	if ( (image(x,y) < minVal) || (image(x,y) > maxVal) )
		return 0;

	image(x,y) = value;

	int val = 1;
	val += myFloodFill(image, x-1, y, value, minVal, maxVal);
	val += myFloodFill(image, x, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y, value, minVal, maxVal);
	val += myFloodFill(image, x, y+1, value, minVal, maxVal);

	val += myFloodFill(image, x-1, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x-1, y+1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y+1, value, minVal, maxVal);

	return val;
	}

void FilterSeparable(const QVImage<sFloat, 1> &image, QVImage<sFloat, 1> &dest,
	const QVVector &rowFilter, const QVVector &colFilter, const QPoint &destROIOffset)
	{
	const uInt cols = image.getCols(), rows = image.getRows();
	QVImage<sFloat> rowFiltered(cols, rows);
	FilterRow(image, rowFiltered, rowFilter);
	FilterColumn(rowFiltered, dest, colFilter, destROIOffset);
	}

////////////////////////////

#include <qvip/qvimagefeatures/qvcomponenttree.h>

void pruneLowComponentTreeAux(QVImage<uChar> &image, QVComponentTree &componentTree, uInt minArea, uInt node, uInt validThreshold)
	{
	Q_ASSERT(componentTree.area(node)[componentTree.firstThreshold(node)] != 0);
	Q_ASSERT(componentTree.area(node)[componentTree.lastThreshold(node)] != 0);
	Q_ASSERT(validThreshold >= componentTree.lastThreshold(node));

	bool prune = false;
	int lastValidThreshold = validThreshold;

	// Here we decide if this node should be directly pruned
	// or if there's any sub-node we should prune
	for (int threshold = componentTree.lastThreshold(node); threshold >= componentTree.firstThreshold(node) && !prune; threshold--)
		if (componentTree.area(node)[threshold] > 0)
			{
			if (componentTree.area(node)[threshold] < minArea)
				prune = true;
			else
				lastValidThreshold = threshold;
			}

	// We prune node, or get on with it's childrens
	if (prune)
		myFloodFill(image, componentTree.seedX(node), componentTree.seedY(node), lastValidThreshold, 0, lastValidThreshold-1);
	else
		for (uInt child = componentTree.firstChild(node); child != NULL_NODE; child = componentTree.nextSibling(child))
			pruneLowComponentTreeAux(image, componentTree, minArea, child, lastValidThreshold);
	}

void pruneHighComponentTreeAux(QVImage<uChar> &image, QVComponentTree &componentTree, uInt minArea, uInt node, uInt validThreshold)
	{
	Q_ASSERT(componentTree.area(node)[componentTree.firstThreshold(node)] != 0);
	Q_ASSERT(componentTree.area(node)[componentTree.lastThreshold(node)] != 0);
	Q_ASSERT(validThreshold >= componentTree.lastThreshold(node));

	bool prune = false;
	int lastValidThreshold = validThreshold;

	// Here we decide if this node should be directly pruned
	// or if there's any sub-node we should prune
	for (int threshold = componentTree.lastThreshold(node); threshold >= componentTree.firstThreshold(node) && !prune; threshold--)
		if (componentTree.area(node)[threshold] > 0)
			{
			if (componentTree.area(node)[threshold] < minArea)
				prune = true;
			else
				lastValidThreshold = threshold;
			}

	// We prune node, or get on with it's childrens
	if (prune)
		myFloodFill(image, componentTree.seedX(node), componentTree.seedY(node), 255-lastValidThreshold, 255-lastValidThreshold+1, 255-0);
	else
		for (uInt child = componentTree.firstChild(node); child != NULL_NODE; child = componentTree.nextSibling(child))
			pruneHighComponentTreeAux(image, componentTree, minArea, child, lastValidThreshold);
	}

void FilterPruneComponentTreeSmallRegions(QVImage<uChar> &image, QVComponentTree &componentTree, uInt minArea)
	{
	qDebug() << "pruneRegions()";
	if (componentTree.isInverseTree())
		{
		if(componentTree.area(componentTree.rootNode())[componentTree.lastThreshold(componentTree.rootNode())] > minArea)
			pruneHighComponentTreeAux(image, componentTree, minArea, componentTree.rootNode(), componentTree.lastThreshold(componentTree.rootNode()));
		}
	else	{
		if(componentTree.area(componentTree.rootNode())[componentTree.lastThreshold(componentTree.rootNode())] > minArea)
			pruneLowComponentTreeAux(image, componentTree, minArea, componentTree.rootNode(), componentTree.lastThreshold(componentTree.rootNode()));
		}

	qDebug() << "pruneRegions() <~ return";
	}

#include <qvmath/qvdisjointset.h>
#include <qvmath/qvvector.h>

QList<QPointF> GetMaximalResponsePoints1(const QVImage<sFloat> &cornerResponseImage, const double threshold)
	{
	const QRect ROI = cornerResponseImage.getROI();
	const int sizeMax = 2;
	const int cols = cornerResponseImage.getCols(), rows = cornerResponseImage.getRows();
	const int	x0 = ROI.x() + sizeMax,
			y0 = ROI.y() + sizeMax,
			x1 = ROI.x() + ROI.width() - sizeMax,
			y1 = ROI.y() + ROI.height() - sizeMax;

	QVImage<uChar> binaryResponseImage(cols, rows);
	QVIMAGE_INIT_READ(sFloat,cornerResponseImage);
	QVIMAGE_INIT_WRITE(uChar,binaryResponseImage);

	// 1. Get image with local maximums
	Set(0, binaryResponseImage);

	// Recorremos la imagen
	for(int row = y0; row < y1; row++)
		for(int col = x0; col < x1; col++)
			{
			sFloat actual = QVIMAGE_PIXEL(cornerResponseImage, col, row,0);
			if (actual >= threshold)
				{
				QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) = IPP_MAX_8U;
				for (int j = row-sizeMax; (j < row+sizeMax) && (QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) > 0); j++)
					for (int i = col-sizeMax; i < col+sizeMax; i++)
						if ( ((i != col) || (j != row)) && (actual <= QVIMAGE_PIXEL(cornerResponseImage, i, j, 0)) )
							{
							QVIMAGE_PIXEL(binaryResponseImage, col, row, 0) = 0;
							break;
							}
				}
			}

	// 2. Get sorted list of points
	QMap<sFloat, QPointF> sortedPoints;
	for(int row = y0; row < y1; row++)
		for(int col = x0; col < x1; col++)
	//for(uInt row = 0; row < binaryResponseImage.getRows(); row++)
	//	for(uInt col = 0; col < binaryResponseImage.getCols(); col++)
			if (QVIMAGE_PIXEL(binaryResponseImage, col, row,0))
				sortedPoints.insertMulti(QVIMAGE_PIXEL(cornerResponseImage, col, row,0), QPointF(col+2, row+2));

	QList<QPointF> result;
	foreach(sFloat key, sortedPoints.uniqueKeys())
		result += sortedPoints.values(key);

	return result;
	}

QList<QPointF> GetMaximalResponsePoints3(const QVImage<sFloat> &cornerResponseImage, const double threshold)
	{
	static const int	xD[4] = { -1, +0, +1, +0 },
				yD[4] = { +0, -1, +0, +1 };

	const int rows = cornerResponseImage.getRows(), cols = cornerResponseImage.getCols();
	QVDisjointSet disjointSet(cornerResponseImage);
	QVIMAGE_INIT_READ(sFloat,cornerResponseImage);

	// 1. Join in the disjoint set all the neighbour pixels
	for(int row = 1; row < rows-1; row++)
		for(int col = 1; col < cols-1; col++)
			if (QVIMAGE_PIXEL(cornerResponseImage, col, row,0) >= threshold)
				{
				// a. Look for the maximum neighbour for the actual pixel.
				int maxDir = -1;
				sFloat maxVal = QVIMAGE_PIXEL(cornerResponseImage, col, row, 0);

				for (int d = 0; d < 4; d++)
					if ( maxVal < QVIMAGE_PIXEL(cornerResponseImage, col + xD[d], row + yD[d], 0) )
						{
						maxDir = d;
						maxVal = QVIMAGE_PIXEL(cornerResponseImage, col + xD[maxDir], row + yD[maxDir], 0);
						}
	
				// b. If found, join actual pixel to its group.
				if (maxDir > -1)
					disjointSet.unify(col, row, col + xD[maxDir], row + yD[maxDir]);
				}

	// 2. Create sets for all the sets of pixels with more than one element.
	QMap<int, QVVector> hotPoints;
	for(int row = 1; row < rows-1; row++)
		for(int col = 1; col < cols-1; col++)
			{
			const sFloat pixelValue = QVIMAGE_PIXEL(cornerResponseImage, col, row, 0);
			if (QVIMAGE_PIXEL(cornerResponseImage, col, row, 0) >= threshold)
				{
				QVVector v(3);
				v[0] = pixelValue * col;
				v[1] = pixelValue * row;
				v[2] = pixelValue;
				const int setIndex = disjointSet.find(col, row);
				if (hotPoints.contains(setIndex))
					hotPoints[setIndex] += v;
				else
					hotPoints.insert(setIndex, v);
				}
			}

	// 3. Get sorted list of points
	QMap<sFloat, QPointF> sortedPoints;
	foreach(int index, hotPoints.keys())
		{
		QVVector v = hotPoints[index];
		sortedPoints.insertMulti(v[2], QPointF(v[0]/v[2]+2, v[1]/v[2]+2));
		}

	QList<QPointF> result;
	foreach(sFloat key, sortedPoints.uniqueKeys())
		foreach (QPointF point, sortedPoints.values(key))
		result.append(point);

	return result;
	}


////////////////////////////

#define DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(TYPE, C)								\
void FilterEqualizeHistogram(const QVImage<TYPE,C> &image, QVImage<TYPE,C> &equalized, const QPoint &destROIOffset)	\
	{														\
	uInt	rows = image.getRows(), cols = image.getCols();		\
	TYPE	maxVal, minVal;						\
									\
	Max(image,maxVal);						\
	Min(image,minVal);						\
									\
	QVImage<TYPE,C> temp(cols, rows), result(cols, rows);		\
	SubC(image, minVal, temp);					\
	MulC(temp, 255/(maxVal-minVal), result, 1, destROIOffset);	\
	equalized = result;						\
	}

DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(uChar,1);

#define DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM2(TYPE, C)								\
void FilterEqualizeHistogram(const QVImage<TYPE,C> &image, QVImage<TYPE,C> &equalized, const QPoint &destROIOffset)	\
	{														\
	uInt	rows = image.getRows(), cols = image.getCols();		\
	TYPE	maxVal, minVal;						\
									\
	Max(image,maxVal);						\
	Min(image,minVal);						\
									\
	QVImage<TYPE,C> temp(cols, rows), result(cols, rows);		\
	SubC(image, minVal, temp);					\
	MulC(temp, 255/(maxVal-minVal), result, destROIOffset);		\
	equalized = result;						\
	}
DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM2(sFloat,1);



