/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvippworkers.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <QVWorker>
#include <qvipp.h>
#include <QVIndexedStringList>

#ifndef QVIPPWORKERS_H
#define QVIPPWORKERS_H

/*!
@brief Provides a full list of the names of all IPP workers

@returns a list of QString objects, corresponding to the names of the IPP workers
@see newQVIPPWorker
@ingroup qvippworkers
*/
const QStringList getQVIPPWorkersNames();

/*!
@brief Creates an IPP worker object, given its class name

This function can be used as a worker object factory.
Given the class name of an IPP worker included in the QVision framework, this function returns a pointer to a newly created worker of that type.

Any QVision application can dynamically obtain the full list of class names of IPP workers included in the framework with the function @ref getQVIPPWorkersNames.

@param type name of the type of the worker
@param name name which will be asigned to the new worker object
@returns a pointer to a new worker object, or NULL if no ipp worker class exists in the QVision
@ingroup qvippworkers
*/
QVPropertyContainer *newQVIPPWorker(QString type, QString name);

char *QVIPPWorkerName(QVPropertyContainer *cont);

/*!
@brief Worker which applies IPP function CopyP3C3
@ingroup qvippworkers
*/
template <typename TYPE> class QVIPPCopyP3C3Worker: public QVWorker
	{
	public:
		QVIPPCopyP3C3Worker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function CopyC3P3
@ingroup qvippworkers
*/
template <typename TYPE> class QVIPPCopyC3P3Worker: public QVWorker
	{
	public:
		QVIPPCopyC3P3Worker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function Dilate3x3
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPDilate3x3Worker: public QVWorker
	{
	public:
		QVIPPDilate3x3Worker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function Erode3x3
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPErode3x3Worker: public QVWorker
	{
	public:
		QVIPPErode3x3Worker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function ComputeThreshold_Otsu
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPComputeThreshold_OtsuWorker: public QVWorker
	{
	public:
		QVIPPComputeThreshold_OtsuWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function AddRandUniform_Direct
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPAddRandUniform_DirectWorker: public QVWorker
	{
	private:
		unsigned int seed;

	public:
		QVIPPAddRandUniform_DirectWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function FilterBox
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPFilterBoxWorker: public QVWorker
	{
	public:
		QVIPPFilterBoxWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function FilterMedian
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPFilterMedianWorker: public QVWorker
	{
	public:
		QVIPPFilterMedianWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function FilterMax
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPFilterMaxWorker: public QVWorker
	{
	public:
		QVIPPFilterMaxWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function FilterMin
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPFilterMinWorker: public QVWorker
	{
	public:
		QVIPPFilterMinWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function FilterGauss
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPFilterGaussWorker: public QVWorker
	{
	public:
		QVIPPFilterGaussWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function FilterHipass
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPFilterHipassWorker: public QVWorker
	{
	public:
		QVIPPFilterHipassWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function Add
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPAddWorker: public QVWorker
	{
	public:
		QVIPPAddWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function AbsDiff
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPAbsDiffWorker: public QVWorker
	{
	public:
		QVIPPAbsDiffWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function AddC
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPAddCWorker: public QVWorker
	{
	public:
		QVIPPAddCWorker(QString name);
		void iterate();
	};

/*!
@brief Worker which applies IPP function CompareC
@ingroup qvippworkers
*/
template <typename TYPE, int C> class QVIPPCompareCWorker: public QVWorker
	{
	public:
		QVIPPCompareCWorker(QString name);
		void iterate();
	};

#endif
