/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvfunctionminimizer.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVFUNCTIONMINIMIZER_H
#define QVFUNCTIONMINIMIZER_H

#include <qvmath.h>
#include <QVVector>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_min.h>

/*!
@class QVFunctionMinimizer qvfunctionminimizer.h qvmath/qvfunctionminimizer.h
@brief Wrapper class for GSL function minimization.

This is a baseclass to create function minimizator objects. It is based on the GSL functionality for that matter.

To use this class, a subclass must be created containing the function to be optimized. Method @ref function should be redefined in the subclass with the code for that function. An example subclass for QVFunctionMinimization usage follows:

@code
class FMinimizer: public QVFunctionMinimizer
	{
	public:
		fMinimizer(): QVFunctionMinimizer()	{}
		double function(double value)
			{ return f(value); }	// Or whatever the function is.
	}
@endcode

The following code ilustrates the usage of this class to perform optimization for function <i>f</i>

@code
[...]
// We create the minimization object
FMinimizer fMinimizer;

// and start a minimization search of 100 iterations 
if (fMinimizer.iterate(100))
	std::cout	<< "The search found a minimum at value "
			<< fMinimizer.getMinimum() << std::endl;
else
	std::cout	<< "No minimum could be found." << std::endl;
[...]
@endcode

Parameter <i>useGoldenSection</i> in the constructor indicates which of the two minimization algorithms implemented in the GSL will be used to perform the minimization: the <a href="http://www.math.uni-bremen.de/techdoc/software/GSL/gsl-ref_31.html#SEC417">Golden Section algorithm</a>, or the <a href="http://www.math.uni-bremen.de/techdoc/software/GSL/gsl-ref_31.html#SEC417">Brent minimization algorithm</a>.

@ingroup qvfunctionoptimization
*/
class QVFunctionMinimizer
	{
	private:
		int iterations;
		bool useGoldenSection;
		double actual, minimum, maximum;

		static double static_function (const double v, void *functionMinimizerPtr);

	public:	
		/// @brief Constructor for the class
		///
		/// @param useGolenSection indicates wether the minimization algorithm used should be the Golden Section or the Brent algorithm.
		QVFunctionMinimizer(const bool useGoldenSection = true):
			useGoldenSection(useGoldenSection), actual(0), minimum(-1), maximum(1)	{ }

		/// @brief Function to be optimized
		///
		/// This  method should be implemented by the subclass
		/// containing the function to be minimized.
		///
		/// @param value value to evaluate the function.
		virtual double function(const double value) = 0;

		/// @brief Specifies the search starting value.
		///
		/// The search will start looking for the minimum around it.
		/// By default this value is set to 0 when the minimizer object
		/// is constructed.
		///
		/// @param value the starting search value.
		void setStartingValue(const double value)			{ actual = value; }

		/// @brief Specifies the search range.
		///
		/// The search will constrain to the given range.
		/// By default this range is [-1,1].
		///
		/// @param value the starting search value.
		void setSearchRange(const double min, const double max )	{ minimum = min; maximum = max; }

		/// @brief Returns the minimum obtained in the last search
		///
		/// After a successful search started with function @ref iterate,
		/// this method return the minimum value obtained in it, if
		/// the search was successful.
		///
		/// @return minimum value obtained in the last search.
		const double & getMinimum()	const	{ return actual; }

		/// @brief Number of iterations elapsed in the last search
		///
		/// This will return the number of iterations which took to the
		/// last search (started with function @ref iterate) to find a
		/// minimum, or stopping, if none was found.
		///
		/// @return number of iterations elapse in the last search.
		const int & getIterations()	const	{ return iterations; }

		/// @brief Launchs a minimization search
		///
		/// This method will start a new search.
		///
		/// @param maxIteration maximum number of steps to perform in the search.
		/// @param maxError maximal relative error admisible for the solution. The minimization will stop when reached.
		/// @returns true if a minimum was found, else false.
		bool iterate(const int maxIterations = 100, const double maxError = 0.00001);
	};

#endif

