/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmatrix.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVMATRIX_H
#define QVMATRIX_H

#include <math.h>
#include <gsl/gsl_blas.h>
#include <iostream>
#include <QVQuaternion>
#include <qvmath/qvblasdatabuffer.h>

/*!
@class QVMatrix qvmatrix.h qvmath/qvmatrix.h
@brief Implementation of numerical matrices.

@ingroup qvmath
*/
class QVMatrix
	{
	public:
		// Constructors

		/// @brief Default constructor
		///
		/// By default, it constructs a 1x1 matrix.
		///
		QVMatrix();

		/// @brief Copy constructor.
		///
		/// Creates a matrix object copying the content from a given one.
		/// @param matrix matrix to be copied.
		QVMatrix(const QVMatrix &matrix);

		/// @brief Dimensional constructor
		///
		/// Creates a new matrix of the specified dimensions.
		///
		/// @param rows number of rows for the new matrix.
		/// @param cols number of rows for the new matrix.
		QVMatrix(const int rows, const int cols);

		/// @brief Rotation matrix constructor
		///
		/// Creates a rotation matrix equivalent to a quaternion.
		///
		/// @param quaternion quaternion equivalent to the rotation.
		QVMatrix(const QVQuaternion &quaternion);

		/// @brief Vector to matrix constructor
		///
		/// Creates a row or column matrix from a vector.
		///
		/// @param vector vector to convert to matrix type.
		/// @param rowVector if true, creates a row matrix from the vector, else creates a column matrix from it.
		QVMatrix(const QVVector &vector, const bool rowVector = true);

		/// @brief Constructor from a list of QVVectors
		///
		/// Creates a new matrix from a list of row QVVectors.
		///
		/// @param vectorList list of row vectors.
		QVMatrix(const QList<QVVector> &vectorList);

		/// @brief Constructor from a list of QVectors
		///
		/// Creates a new matrix from a list of row QVectors.
		///
		/// @param vectorList list of row vectors.
		QVMatrix(const QList< QVector<double> > &vectorList);

		/// @brief Copy constructor for gsl_matrix
		///
		/// @param matrix matrix to be copied
		QVMatrix(const gsl_matrix *matrix);

		/// @brief Constructor from a list of QPointF
		///
		/// Creates a new matrix from a list of points
		///
		/// @param pointList list of points.
		QVMatrix(const QList<QPointF> &pointList);

		/// @brief Convert to a list of QVVectors operator
		///
		// Creates a new list of row QVVectors from a matrix
		operator QList<QVVector> () const
			{
			QList<QVVector> list;
			for (int n = 0; n < getRows(); n++)
				list.append(getRow(n));
			return list;
			}

		/// @brief Convert to a list of QVectors operator
		///
		// Creates a new list of row QVectors from a matrix
		operator QList< QVector<double> > () const
			{
			QList< QVector<double> > list;
			for (int n = 0; n < getRows(); n++)
				list.append(getRow(n));
			return list;
			}

		/// @brief Convert to a gsl_matrix * operator
		///
		// Creates a new gsl_matrix * from a matrix
		operator gsl_matrix * () const
			{
			gsl_matrix *result = gsl_matrix_alloc(rows, cols);
			for(int i = 0; i < rows; i++)
				for(int j = 0; j < cols; j++)
					gsl_matrix_set(result, i, j, operator()(i, j));
			return result;
			}

		/// @brief Copy operator
		///
		/// @param matrix matrix to be copied.
		QVMatrix & operator=(const QVMatrix &matrix);

		// Matrix-matrix operators

		/// @brief Matrix-matrix equality operator
		///
		/// @param matrix term for the compare operation.
		/// @return true if the given matrix has same dimensions and equal content of the data buffer, else false.
		/// @see equals
		bool operator==(const QVMatrix &matrix) const			{ return equals(matrix);	};

		/// @brief Matrix-matrix unequality operator
		///
		/// @param matrix term for the compare operation.
		/// @return false if the given matrix has same dimensions and equal content of the data buffer, else true.
		/// @see equals
		bool operator!=(const QVMatrix &matrix) const			{ return !equals(matrix);	};

		/// @brief Matrix-matrix product operator
		///
		/// @param matrix factor for the product operation.
		/// @see matrixProduct
		QVMatrix operator*(const QVMatrix &matrix) const		{ return dotProduct(matrix); 	};

		/// @brief Matrix-matrix division
		///
		/// The following expression:
		///
		///	X = A / B
		///
		/// is equivalent to this one
		///
		///	X = A * pseudoInverse(B)
		///
		/// With the difference that the pseudoinverse matrix for <i>B</i> is not directly computed. Instead, the
		/// matrix equation:
		///
		///	X * B = A
		///
		/// is solved for the <i>X</i> term. This leads to more stable and efficient results than using the pseudo inverse.
		///
		/// this operator gets X from:
		///
		///	X = A / B	->	X * B = A	->	B' * X' = A'
		///
		/// using a linear system resolution based on a LU decomposition over A.
		///
		/// @param matrix factor for the operation.
		/// @see matrixDivide(const QVMatrix &) const
		QVMatrix operator/(const QVMatrix &matrix) const		{ return matrixDivide(matrix); 	};

		/// @brief Matrix-matrix addition operator
		///
		/// @param matrix term for the operation.
		/// @see addition
		QVMatrix operator+(const QVMatrix &matrix) const		{ return addition(matrix); 	};

		/// @brief Matrix-matrix substraction operator
		///
		/// @param matrix subtrahend for the operation.
		/// @see substract
		QVMatrix operator-(const QVMatrix &matrix) const		{ return substract(matrix); 	};

		/// @brief Matrix negation
		///
		/// Obtains a matrix negating all the elements of the original.
		///
		/// @see substract
		QVMatrix operator-() const					{ return operator*(-1); 	};

		// Matrix-scalar operators

		/// @brief Matrix-scalar product
		///
		/// @param value factor for the operation.
		/// @see scalarProduct
		QVMatrix operator*(const double value) const			{ return scalarProduct(value); 	};

		/// @brief Matrix-scalar division operator
		///
		/// @param value divider for the operation.
		/// @see scalarDivide
		QVMatrix operator/(const double value) const			{ return scalarDivide(value); 	};

		// Matrix-vector operators

		/// @brief Matrix-vector product
		///
		/// @param vector vector to multiply by.
		QVVector operator*(QVVector &vector) const;

		// Operation methods

		/// @brief Cell access operator
		///
		/// @param row main index value for data access
		/// @param col secondary index value for data access
		inline double &operator()(const int row, const int col)
			{ return data->getWriteData()[row*cols + col]; }

		/// @brief Overloaded version of cell access operator
		///
		/// @param row main index value for data access
		/// @param col secondary index value for data access
		inline const double operator()(const int row, const int col) const
			{ return data->getReadData()[row*cols + col]; }

		/// @brief Gets the size of the data array.
		///
		/// This returns the number of elements in the matrix.
		/// @return the product of the sizes of all the indexes in the matrix.
		const int getDataSize()			const	{ return cols*rows; }

		/// @brief Gets a read-only reference to the data buffer of the matrix
		///
		/// @return a constant pointer to the data buffer of the matrix
		const double *getReadData()		const	{ return data->getReadData(); }

		/// @brief Gets a reference of the data buffer of the matrix for read and write accesses
		///
		/// @return a pointer to the data buffer of the matrix
		double *getWriteData()				{ return data->getWriteData(); }

		/// @brief Change the order of the indexes in the matrix.
		///
		/// This function reorders the indexes of the matrix.
		QVMatrix transpose() const;

		/// @brief Sets the value of the elements of the matrix
		///
		/// @param value value to set elements in the matrix
		void set(const double value);

		// Matrix-matrix operations

		/// @brief Matrix-matrix equality operator
		///
		/// @param matrix term for the compare operation.
		/// @return true if the given matrix has same dimensions and equal content of the data buffer, else false.
		/// @see operator==(const QVMatrix &) const
		/// @see operator!=(const QVMatrix &) const
		bool equals(const QVMatrix &matrix) const;

		/// @brief Matrix-matrix product
		///
		/// @param matrix factor for the operation.
		/// @see operator*(const QVMatrix &) const
		QVMatrix dotProduct(const QVMatrix &matrix) const;

		/// @brief Matrix-matrix division
		///
		/// This function solves the <i>X</i> term from the following matrix equation:
		///	<center>X A = B</center>
		/// where <i>A</i> and <i>B</i> are respectively the actual matrix, and the one given by parameter.
		/// This is equivalent to using the following expression for <i>X</i>:
		///	<center>X = B A<sup>-1</sup></center>
		/// but with a faster method returning a more stable result.
		///
		/// @param matrix factor for the operation.
		/// @see operator/(const QVMatrix &) const
		QVMatrix matrixDivide(const QVMatrix &matrix) const;

		/// @brief Computes the inverse of the matrix.
		QVMatrix inverse() const;

		/// @brief Computes the determinant of the matrix.
		double det() const;

		/// @brief Obtains the mean vector for the column vectors in the matrix
		QVVector meanCol() const;

		/// @brief Matrix-matrix addition
		///
		/// @param matrix operand for the add operation.
		/// @see operator+(const QVMatrix &) const
		QVMatrix addition(const QVMatrix &matrix) const;

		/// @brief Matrix-matrix substraction
		///
		/// @param matrix subtrahend for the operation.
		/// @see operator-(const QVMatrix &) const
		QVMatrix substract(const QVMatrix &matrix) const;

		// Matrix-scalar operations

		/// @brief Matrix-scalar division
		///
		/// @param value scalar value to use in the operation.
		/// @see operator/(const double) const
		QVMatrix scalarDivide(const double value) const;

		/// @brief Matrix-scalar product
		///
		/// @param value factor for the operation.
		/// @see operator*(const double) const
		QVMatrix scalarProduct(const double value) const;

		/// @brief Gets the <a href="http://en.wikipedia.org/wiki/Norm_%28mathematics%29">norm<sub>2</sub></a> for matrix.
		///
		/// The resulting value will be the square root of the following sumatory:
		///
		/// \f$ \|A\|_2 := \sqrt{ \sum |a_{ij}|^2 } \f$
		///
		/// where \f$ a_{ij} \f$ are the elements of the matrix \f$ A \f$.
		/// @return norm-2 of the matrix.
		double norm2() const;

		/// @brief Homogeneous normalization for rows of the matrix
		///
		/// @todo explain homogeneous normalization
		///
		/// @return norm-2 of the matrix.
		QVMatrix rowHomogeneousNormalize() const;

		/// @brief Get width of the matrix.
		///
		/// @return number of columns for the matrix.
		const int getCols() const	{ return cols; }

		/// @brief Get height of the matrix.
		///
		/// @return number of rows for the matrix.
		const int getRows() const	{ return rows; }

		// Access operations

		/// @brief Gets a row of the matrix
		///
		/// @param row row number, a value within [0, r-1], for a matrix with 'r' rows
		/// @returns a vector containing the values of the row
		const QVVector getRow(const int row) const;

		/// @brief Sets a row of the matrix
		///
		/// @param row row number, a value within [0, r-1], for a matrix with 'r' rows
		/// @param vector a vector containing the values for the row
		void setRow(const int row, QVVector vector);

		/// @brief Sets a row of the matrix
		///
		/// @param row row number, a value within [0, r-1], for a matrix with 'r' rows
		/// @param vector a vector containing the values for the row
		void setRow(const int row, QVector<double> vector);

		/// @brief Gets a column of the matrix
		///
		/// @param col column number, a value within [0, c-1], for a matrix with 'c' columns
		/// @returns a vector containing the values of the column
		const QVVector getCol(const int col) const;

		/// @brief Sets a column of the matrix
		///
		/// @param col column number, a value within [0, c-1], for a matrix with 'c' columns
		/// @param vector a vector containing the values for the column
		void setCol(const int col, QVVector vector);

		/// @brief Gets a submatrix from a matrix
		///
		/// @param firstCol first column number for the submatrix
		/// @param firstRow first row number for the submatrix
		/// @param lastCol last column number for the submatrix
		/// @param lastRow last row number for the submatrix
		/// @returns a submatrix of size (lastRow - firstRow + 1)x(lastCol - firstCol + 1) containing the corresponding elements from the original matrix.
		const QVMatrix getSubmatrix(const int firstRow, const int lastRow, const int firstCol, const int lastCol) const;

		// Misc matrices

		/// @brief Creates an identity matrix
		///
		/// @param size number of cols and rows for the identity matrix
		static QVMatrix identity(const int size);

		/// @brief Creates a random matrix
		///
		/// Matrix with elements values from 0 to 1.
		///
		/// @param rows rows of the matrix
		/// @param cols cols of the matrix
		static QVMatrix random(const int rows, const int cols);

		/// @brief Creates a diagonal matrix
		///
		/// Matrix with elements values from 0 to 1.
		///
		/// @param diagonalVector vector containing the diagonal values
		static QVMatrix diagonal(const QVVector &diagonalVector);

		/// @brief Create a planar rotation matrix
		///
		/// @param angle angle to rotate around coordinate center
		static QVMatrix rotationMatrix(const double angle);

		/// @brief Create a 3D traslation matrix.
		///
		/// @param x increment in X axis
		/// @param y increment in Y axis
		static QVMatrix traslationMatrix(const double x, const double y);

		/// @brief Create a 3D translation matrix.
		///
		/// @param zoom scale
		static QVMatrix scaleMatrix(const double zoom);

		/// @brief Create a rotation matrix for 3D-X axis
		///
		/// @param angle angle to rotate around X axis in radians
		static QVMatrix rotationMatrix3dXAxis(const double angle);

		/// @brief Create a rotation matrix for 3D-Y axis
		///
		/// @param angle angle to rotate around Y axis in radians
		static QVMatrix rotationMatrix3dYAxis(const double angle);

		/// @brief Create a rotation matrix for 3D-Z axis
		///
		/// @param angle angle to rotate around Z axis in radians
		static QVMatrix rotationMatrix3dZAxis(const double angle);

		/// @brief Create a 3D translation matrix.
		///
		/// @param x increment in X axis
		/// @param y increment in Y axis
		/// @param z increment in Z axis
		static QVMatrix traslationMatrix3d(const double x, const double y, const double z);

	private:
		int cols, rows;
		QSharedDataPointer< QBlasDataBuffer > data;
	};

/// @todo detailed description should follow
/// @ingroup qvmath
std::ostream& operator << ( std::ostream &os, const QVMatrix &matrix );

/// @todo detailed description should follow
/// @ingroup qvmath
std::istream& operator >> ( std::istream &is, QVMatrix &matrix );

/// @todo detailed description should follow
/// @ingroup qvmath
uint qHash(const QVMatrix &matrix);

#include <QMetaType>
Q_DECLARE_METATYPE(QVMatrix);

#endif

